# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pprint
import re
import sys
import unittest

import unicodecsv as csv

from nose.plugins.attrib import attr

from hgvs.exceptions import HGVSError
import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.variant
import hgvs.variantmapper


def gxp_file_reader(fn):
    rdr = csv.DictReader(open(fn, 'r'), delimiter=str('\t'))
    for rec in rdr:
        if rec['id'].startswith('#'):
            continue
        yield rec


@attr(tags=["mapping"])
class Test_VariantMapper(unittest.TestCase):
    def setUp(self):
        self.hdp = hgvs.dataproviders.uta.connect()
        self.hm = hgvs.variantmapper.VariantMapper(self.hdp)
        self.hp = hgvs.parser.Parser()

    # ZCCHC3 -- one exon, + strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = 'NM_033089.6';
    # ┌────────┬────────────┬─────────┬─────────────┬──────────────┬─────────────┬─────────────┬───────────┐
    # │  hgnc  │ alt_strand │ n_exons │    tx_ac    │    alt_ac    │  s_cigars   │ cds_start_i │ cds_end_i │
    # ├────────┼────────────┼─────────┼─────────────┼──────────────┼─────────────┼─────────────┼───────────┤
    # │ ZCCHC3 │          1 │       1 │ NM_033089.6 │ NC_000020.10 │ 484=3I2275= │          24 │      1236 │
    # └────────┴────────────┴─────────┴─────────────┴──────────────┴─────────────┴─────────────┴───────────┘
    def test_ZCCHC3_dbSNP(self):
        for rec in gxp_file_reader('tests/data/gcp/ZCCHC3-dbSNP.tsv'):
            self._test_gxp_mapping(rec)

    # ORAI1 -- two exons, + strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = 'NM_032790.3';
    # ┌───────┬────────────┬─────────┬─────────────┬──────────────┬──────────────────┬─────────────┬───────────┐
    # │ hgnc  │ alt_strand │ n_exons │    tx_ac    │    alt_ac    │     s_cigars     │ cds_start_i │ cds_end_i │
    # ├───────┼────────────┼─────────┼─────────────┼──────────────┼──────────────────┼─────────────┼───────────┤
    # │ ORAI1 │          1 │       2 │ NM_032790.3 │ NC_000012.11 │ 319=6I177=;1000= │         193 │      1099 │
    # └───────┴────────────┴─────────┴─────────────┴──────────────┴──────────────────┴─────────────┴───────────┘
    def test_ORAI1_dbSNP(self):
        for rec in gxp_file_reader('tests/data/gcp/ORAI1-dbSNP.tsv'):
            self._test_gxp_mapping(rec)

    # FOLR3 -- multiple exons, + strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = 'NM_000804.2';
    # ┌───────┬────────────┬─────────┬─────────────┬─────────────┬──────────────────────────────┬─────────────┬───────────┐
    # │ hgnc  │ alt_strand │ n_exons │    tx_ac    │   alt_ac    │           s_cigars           │ cds_start_i │ cds_end_i │
    # ├───────┼────────────┼─────────┼─────────────┼─────────────┼──────────────────────────────┼─────────────┼───────────┤
    # │ FOLR3 │          1 │       5 │ NM_000804.2 │ NC_000011.9 │ 44=;174=;150=2D37=;136=;304= │          50 │       788 │
    # └───────┴────────────┴─────────┴─────────────┴─────────────┴──────────────────────────────┴─────────────┴───────────┘
    def test_FOLR3_dbSNP(self):
        # TODO: CORE-158: g-to-c mapped insertions have incorrect interval bounds
        for rec in gxp_file_reader('tests/data/gcp/FOLR3-dbSNP.tsv'):
            self._test_gxp_mapping(rec)

    # ADRA2B -- one exon, - strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = 'NM_000682.5';
    # ┌────────┬────────────┬─────────┬─────────────┬──────────────┬─────────────┬─────────────┬───────────┐
    # │  hgnc  │ alt_strand │ n_exons │    tx_ac    │    alt_ac    │  s_cigars   │ cds_start_i │ cds_end_i │
    # ├────────┼────────────┼─────────┼─────────────┼──────────────┼─────────────┼─────────────┼───────────┤
    # │ ADRA2B │         -1 │       1 │ NM_000682.5 │ NC_000002.11 │ 891=9D2375= │           0 │      1353 │
    # └────────┴────────────┴─────────┴─────────────┴──────────────┴─────────────┴─────────────┴───────────┘
    def test_ADRA2B_dbSNP(self):
        for rec in gxp_file_reader('tests/data/gcp/ADRA2B-dbSNP.tsv'):
            self._test_gxp_mapping(rec)

    # JRK -- multiple exons, - strand
    # reece@[local]/uta_dev=> select hgnc,alt_strand,n_exons,tx_ac,alt_ac,s_cigars,cds_start_i,cds_end_i from bermuda.bermuda_data_mv where tx_ac = 'NM_001077527.1';
    # ┌──────┬────────────┬─────────┬────────────────┬──────────────┬───────────────────────┬─────────────┬───────────┐
    # │ hgnc │ alt_strand │ n_exons │     tx_ac      │    alt_ac    │       s_cigars        │ cds_start_i │ cds_end_i │
    # ├──────┼────────────┼─────────┼────────────────┼──────────────┼───────────────────────┼─────────────┼───────────┤
    # │ JRK  │         -1 │       3 │ NM_001077527.1 │ NC_000008.10 │ 52=;1844=2I199=;1483= │         514 │      2185 │
    # └──────┴────────────┴─────────┴────────────────┴──────────────┴───────────────────────┴─────────────┴───────────┘
    def test_JRK_dbSNP(self):
        # TODO: CORE-157: del26 on -1 strands gets reverse complemented as del62
        for rec in gxp_file_reader('tests/data/gcp/JRK-dbSNP.tsv'):
            self._test_gxp_mapping(rec)

    def test_NEFL_dbSNP(self):
        for rec in gxp_file_reader('tests/data/gcp/NEFL-dbSNP.tsv'):
            self._test_gxp_mapping(rec)

    def test_DNAH11_hgmd(self):
        for rec in gxp_file_reader('tests/data/gcp/DNAH11-HGMD.tsv'):
            self._test_gxp_mapping(rec)

    def test_DNAH11_dbSNP_NM_003777(self):
        for rec in gxp_file_reader('tests/data/gcp/DNAH11-dbSNP-NM_003777.tsv'):
            self._test_gxp_mapping(rec)

    def test_DNAH11_dbSNP_NM_001277115(self):
        for rec in gxp_file_reader('tests/data/gcp/DNAH11-dbSNP-NM_001277115.tsv'):
            self._test_gxp_mapping(rec)

    @attr(tags=["regression"])
    def test_regression(self):
        for rec in gxp_file_reader('tests/data/gcp/regression.tsv'):
            self._test_gxp_mapping(rec)

    @attr(tags=["extra"])
    def test_DNAH11_dbSNP_full(self):
        for rec in gxp_file_reader('tests/data/gcp/DNAH11-dbSNP.tsv'):
            self._test_gxp_mapping(rec)

    def test_real(self):
        for rec in gxp_file_reader('tests/data/gcp/real.tsv'):
            self._test_gxp_mapping(rec)

    def test_noncoding(self):
        for rec in gxp_file_reader('tests/data/gcp/noncoding.tsv'):
            self._test_gxp_mapping(rec)

    def _test_gxp_mapping(self, rec):
        """given one record (row) of g, c/n/r, and p (optional) test variants, map
        g->c/n/r, c/n/r->g, and c->p and verify equivalence

        """

        def _rm_del_seq(vs):
            return re.sub(vs, 'del\w+ins', 'delins')

        var_g = self.hp.parse_hgvs_variant(rec['HGVSg'])
        var_x = self.hp.parse_hgvs_variant(rec['HGVSc'])
        var_p = self.hp.parse_hgvs_variant(rec['HGVSp']) if rec['HGVSp'] is not None and rec['HGVSp'] != '' else None

        # g -> x
        if var_x.type == 'c':
            var_x_test = self.hm.g_to_c(var_g, var_x.ac)
        elif var_x.type == 'n':
            var_x_test = self.hm.g_to_n(var_g, var_x.ac)
        self.assertEquals(_rm_del_seq(str(var_x)), _rm_del_seq(str(var_x_test)),
                          msg="%s != %s (%s; HGVSg=%s)" % (str(var_x_test), str(var_x), rec['id'], rec['HGVSg']))

        # c,n -> g
        if var_x.type == 'c':
            var_g_test = self.hm.c_to_g(var_x, var_g.ac)
        elif var_x.type == 'n':
            var_g_test = self.hm.n_to_g(var_x, var_g.ac)
        self.assertEquals(_rm_del_seq(str(var_g)), _rm_del_seq(str(var_g_test)),
                          msg="%s != %s (%s; HGVSc=%s)" % (str(var_g_test), str(var_g), rec['id'], rec['HGVSc']))

        if var_p is not None:
            # c -> p
            hgvs_p_exp = str(var_p)
            var_p_test = self.hm.c_to_p(var_x, var_p.ac)

            if not var_p.posedit.uncertain:
                # if expected value isn't uncertain, strip uncertain from test
                var_p_test.posedit.uncertain = False

            hgvs_p_test = str(var_p_test)

            if re.search('Ter$', hgvs_p_exp):
                # if expected value doesn't have a count, strip it from the test
                hgvs_p_test = re.sub('Ter\d+$', 'Ter', hgvs_p_test)

            self.assertEquals(hgvs_p_exp, hgvs_p_test, msg="%s != %s (%s)" % (hgvs_p_exp, hgvs_p_test, rec['id']))


if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nose.plugins.attrib import attr

from hgvs.exceptions import HGVSValidationError, HGVSUnsupportedOperationError
import hgvs.dataproviders.uta
import hgvs.variantmapper
import hgvs.parser
import hgvs.validator

hdp = hgvs.dataproviders.uta.connect()


class Test_HGVSValidator(unittest.TestCase):
    """Validator wrapper class tests (most testing is handled by the component classes)"""

    @classmethod
    def setUp(cls):
        cls.hp = hgvs.parser.Parser()
        cls.vr = hgvs.validator.Validator(hdp)

    def test_wrapper(self):
        """Test that validator wrapper is working"""
        self.assertTrue(self.vr.validate(self.hp.parse_hgvs_variant('NM_001005405.2:c.6C>A')))


@attr(tags=["quick", "validation"])
class Test_HGVSIntrinsicValidator(unittest.TestCase):
    """Tests for internal validation"""

    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()
        cls.validate_int = hgvs.validator.IntrinsicValidator()

    def test_start_lte_end(self):
        """Test if start position is less <= end position"""
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561662C>T')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561662_36561663insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NM_01234.5:c.76_77insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54_123+55insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54A>T')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NC_012920.1:m.54G>C')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NC_012920.1:m.57_58insC')))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.BASE_RANGE_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561664_36561663A>T'))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.BASE_RANGE_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('NM_000277.1:c.*1_2delAG'))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.OFFSET_RANGE_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+56_123+55A>T'))

    def test_ins_length_is_one(self):
        """Test if insertion length = 1"""
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NC_000007.13:g.36561662_36561663insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.76_77insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54_123+55insT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123-54_123-53insT')))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.INS_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.76_78insTT'))

        with self.assertRaisesRegexp(HGVSValidationError, hgvs.validator.INS_ERROR_MSG):
            self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54_123+56insT'))

    def test_del_length(self):
        """Test if del length agrees with position range"""
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.76_78delACT')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.123+54_123+55delTA')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('AC_01234.5:c.76_78del')))
        self.assertTrue(self.validate_int.validate(self.hp.parse_hgvs_variant('NM_003060.3:c.-91_22del113')))

        # the variants are expected to fail:
        for h in ['AC_01234.5:c.76_78delACTACAT']:
            with self.assertRaises(HGVSValidationError):
                self.validate_int.validate(self.hp.parse_hgvs_variant(h))

        for h in ['NM_032487.4:c.831_*2687del2976']:
            with self.assertRaises(HGVSUnsupportedOperationError):
                self.validate_int.validate(self.hp.parse_hgvs_variant(h))


@attr(tags=["validation"])
class Test_HGVSExtrinsicValidator(unittest.TestCase):
    """Tests for external validation"""

    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()
        cls.validate_ext = hgvs.validator.ExtrinsicValidator(hdp)

    def test_valid_ref(self):
        """Test variants with valid reference seqeuences."""
        hgvs_variants = ['NM_001005405.2:c.6C>A',
                         'NM_001005405.2:c.-38T>A',
                         'NM_001005405.2:c.*3C>G',
                         'NM_001005405.2:c.435_440delCTGCTG',
                         'NM_000187.3:c.457dupG',
                         'NM_000151.3:c.379_380dup2',
                         'NR_002728.3:r.55A>U',
                         'NM_000495.3:c.610_628del19',
                         'NM_000059.3:c.44_45insATT',
                         'NC_000001.10:g.156104193G>A',
                         ]
        for var in [self.hp.parse_hgvs_variant(h) for h in hgvs_variants]:
            self.assertTrue(self.validate_ext.validate(var))

    def test_invalid_ref(self):
        """Test variants with invalid reference sequences."""
        hgvs_variants = ['NM_001005405.2:c.435_440delCTGCT',
                         'NR_002728.3:r.55C>U',
                         'NC_000001.10:g.156104193C>A',
                         ]

        for var in [self.hp.parse_hgvs_variant(h) for h in hgvs_variants]:
            with self.assertRaises(HGVSValidationError):
                self.validate_ext.validate(var)


if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import recordtype

from hgvs.exceptions import HGVSError, HGVSUnsupportedOperationError


class PosEdit(recordtype.recordtype('PosEdit', [('pos', None), ('edit', None), ('uncertain', False)])):
    """
    represents a **simple** variant, consisting of a single position and edit pair
    """

    def __str__(self):
        rv = str(self.edit) if self.pos is None else '{self.pos}{self.edit}'.format(self=self)
        if self.uncertain:
            if self.edit in ['0', '']:
                rv = rv + '?'
            else:
                rv = '(' + rv + ')'
        return rv

    def _set_uncertain(self):
        """sets the uncertain flag to True; used primarily by the HGVS grammar

        :returns: self
        """
        self.uncertain = True
        return self


    def length_change(self, on_error_raise=True):
        """Returns the net length change for this posedit.

        The method for computing the net length change depends on the
        type of variant (dup, del, ins, etc).  The length_change
        method hides this complexity from callers.

        :param hgvs.posedit.PosEdit self: a PosEdit instance
        :param bool on_error_raise: whether to raise an exception on errors 

        :returns: A signed int for the net change in length.  Negative \
        values imply net deletions, 0 implies a balanced insertion and \
        deletion (e.g., SNV), and positive values imply a net \
        insertion.

        :raises HGVSUnsupportedOperationError: When determining the \
        length for this variant type is ill-defined or unsupported.

        There are many circumstances in which the net length change
        cannot be determined, is ill-defined, or is unsupported.  In
        these cases, the result depends on the value of
        `on_error_raise`: when `on_error_raise` is True, an exception
        is raised; when False, the exception is caught and `None` is
        returned.  Callers might wish to pass `on_error_raise=False`
        in list comprehensions to avoid dealing with exceptions.

        """

        try:
            ilen = self.pos._length()
            (del_len, ins_len) = self.edit._del_ins_lengths(ilen)
            return ins_len - del_len
        except HGVSUnsupportedOperationError:
            if on_error_raise:
                raise
            return None



    ## <LICENSE>
    ## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
    ## 
    ## Licensed under the Apache License, Version 2.0 (the "License");
    ## you may not use this file except in compliance with the License.
    ## You may obtain a copy of the License at
    ## 
    ##     http://www.apache.org/licenses/LICENSE-2.0
    ## 
    ## Unless required by applicable law or agreed to in writing, software
    ## distributed under the License is distributed on an "AS IS" BASIS,
    ## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    ## See the License for the specific language governing permissions and
    ## limitations under the License.
    ## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nose.plugins.attrib import attr

import hgvs.edit
import hgvs.location
from hgvs.exceptions import HGVSError


@attr(tags=["quick", "models"])
class Test_Edit(unittest.TestCase):
    def test_NARefAlt_exceptions(self):
        with self.assertRaises(HGVSError):
            edit = str(hgvs.edit.NARefAlt(None, None))

    def test_NARefAlt(self):
        self.assertEqual(str(hgvs.edit.NARefAlt('A', 'A')), '=')
        self.assertEqual(str(hgvs.edit.NARefAlt('A', 'T')), 'A>T')
        self.assertEqual(str(hgvs.edit.NARefAlt('AA', None)), 'delAA')
        self.assertEqual(str(hgvs.edit.NARefAlt(None, 'TT')), 'insTT')
        self.assertEqual(str(hgvs.edit.NARefAlt('AA', 'T')), 'delAAinsT')
        self.assertEqual(str(hgvs.edit.NARefAlt('A', 'TT')), 'delAinsTT')
        # edit types
        self.assertEqual(str(hgvs.edit.NARefAlt('A', 'A').type), 'identity')
        self.assertEqual(str(hgvs.edit.NARefAlt('A', 'T').type), 'sub')
        self.assertEqual(str(hgvs.edit.NARefAlt('AA', None).type), 'del')
        self.assertEqual(str(hgvs.edit.NARefAlt(None, 'TT').type), 'ins')
        self.assertEqual(str(hgvs.edit.NARefAlt('AA', 'T').type), 'delins')
        self.assertEqual(str(hgvs.edit.NARefAlt('A', 'TT').type), 'delins')

    def test_AARefAlt(self):
        self.assertEqual(str(hgvs.edit.AARefAlt('A', 'A')), '=')
        self.assertEqual(str(hgvs.edit.AARefAlt('A', 'T')), 'Thr')
        self.assertEqual(str(hgvs.edit.AARefAlt('AA', None)), 'del')
        self.assertEqual(str(hgvs.edit.AARefAlt(None, 'TT')), 'insThrThr')
        self.assertEqual(str(hgvs.edit.AARefAlt('', 'T')), 'delinsThr')
        self.assertEqual(str(hgvs.edit.AARefAlt('AA', 'T')), 'delinsThr')
        self.assertEqual(str(hgvs.edit.AARefAlt('A', 'TT')), 'delinsThrThr')
        # edit types
        self.assertEqual(str(hgvs.edit.AARefAlt('A', 'A').type), 'identity')
        self.assertEqual(str(hgvs.edit.AARefAlt('A', 'T').type), 'sub')
        self.assertEqual(str(hgvs.edit.AARefAlt('AA', None).type), 'del')
        self.assertEqual(str(hgvs.edit.AARefAlt(None, 'TT').type), 'ins')
        self.assertEqual(str(hgvs.edit.AARefAlt('', 'T').type), 'delins')
        self.assertEqual(str(hgvs.edit.AARefAlt('AA', 'T').type), 'delins')
        self.assertEqual(str(hgvs.edit.AARefAlt('A', 'TT').type), 'delins')

    def test_AASub(self):
        self.assertEqual(str(hgvs.edit.AASub('A', 'T')), 'Thr')
        # edit types
        self.assertEqual(str(hgvs.edit.AASub('A', 'T').type), 'sub')

    def test_AAFs(self):
        self.assertEqual(str(hgvs.edit.AAFs('A', 'T', '6')), 'ThrfsTer6')
        self.assertEqual(str(hgvs.edit.AAFs('A', 'T', '?')), 'ThrfsTer?')
        self.assertEqual(str(hgvs.edit.AAFs('A', 'T', None)), 'ThrfsTer')
        # edit types
        self.assertEqual(str(hgvs.edit.AAFs('A', 'T', '6').type), 'fs')
        self.assertEqual(str(hgvs.edit.AAFs('A', 'T', '?').type), 'fs')
        self.assertEqual(str(hgvs.edit.AAFs('A', 'T', None).type), 'fs')

    def test_AAExt(self):
        self.assertEqual(str(hgvs.edit.AAExt('A', 'V', '*', 10)), 'ValextTer10')
        self.assertEqual(str(hgvs.edit.AAExt('A', 'V', None, -10)), 'Valext-10')
        self.assertEqual(str(hgvs.edit.AAExt('A', None, None, -5)), 'ext-5')
        # edit types
        self.assertEqual(str(hgvs.edit.AAExt('A', 'V', '*', 10).type), 'ext')
        self.assertEqual(str(hgvs.edit.AAExt('A', 'V', None, -10).type), 'ext')
        self.assertEqual(str(hgvs.edit.AAExt('A', None, None, -5).type), 'ext')

    def test_Dup(self):
        self.assertEqual(str(hgvs.edit.Dup()), 'dup')
        self.assertEqual(str(hgvs.edit.Dup('T')), 'dupT')
        # edit types
        self.assertEqual(str(hgvs.edit.Dup().type), 'dup')
        self.assertEqual(str(hgvs.edit.Dup('T').type), 'dup')

    def test_Repeat(self):
        self.assertEqual(str(hgvs.edit.Repeat('CAG', 12, 34)), 'CAG(12_34)')
        # edit types
        self.assertEqual(str(hgvs.edit.Repeat('CAG', 12, 34).type), 'repeat')

    def test_Repeat_exceptions(self):
        with self.assertRaises(HGVSError):
            edit = str(hgvs.edit.Repeat('CAG', 34, 12))

    def test_Inv(self):
        self.assertEqual(str(hgvs.edit.Inv()), 'inv')
        self.assertEqual(str(hgvs.edit.Inv('AGCT')), 'invAGCT')
        self.assertEqual(str(hgvs.edit.Inv('100')), 'inv100')
        # edit types
        self.assertEqual(str(hgvs.edit.Inv().type), 'inv')

    def test_Conv(self):
        start = hgvs.location.BaseOffsetPosition(base=61, offset=-6, datum=hgvs.location.CDS_START)
        end = hgvs.location.BaseOffsetPosition(base=22, datum=hgvs.location.CDS_END)
        pos = hgvs.location.Interval(start=start, end=end)
        self.assertEqual(str(hgvs.edit.Conv('NM_001166478.1', 'c', pos)), 'conNM_001166478.1:c.61-6_*22')
        # edit types
        self.assertEqual(str(hgvs.edit.Conv('NM_001166478.1', 'c', pos).type), 'con')



if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


class HGVSError(Exception):
    pass


class HGVSDataNotAvailableError(HGVSError):
    pass

class HGVSInternalError(HGVSError):
    pass

class HGVSInvalidIntervalError(HGVSError):
    pass

class HGVSInvalidVariantError(HGVSError):
    pass

class HGVSNormalizationError(HGVSError):
    pass

class HGVSParseError(HGVSError):
    pass

class HGVSUnsupportedOperationError(HGVSError):
    pass

class HGVSUsageError(HGVSError):
    """Exception thrown when client/caller has made an invalid request"""
    pass

class HGVSValidationError(HGVSError):
    pass


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



"""Defines a default set of config for use throughout the hgvs
package

Config are read from an ini-format file using configparser. A thin
wrapper on the ConfigParser instance provides *attribute* based
lookups (rather than key), and returns heuristically typed values
(e.g., "True" becomes True).

Although keys are settable, they are stringified on setting and
type-inferred on getting, which means that round-tripping works only
for str, int, and boolean.

>>> import hgvs.config

.. data:: hgvs.config.global_config

  global config, initialized once with defaults

"""

from configparser import ConfigParser, ExtendedInterpolation
from copy import copy
import logging
from pkg_resources import resource_stream
import re


logger = logging.getLogger(__name__)


class Config(object):
    def __init__(self, extended_interpolation=True):
        if extended_interpolation:
            cp = ConfigParser(interpolation=ExtendedInterpolation())
        else:
            cp = ConfigParser()
        cp.optionxform = _name_xform
        self._cp = cp

    def __dir__(self):
        return self._cp.keys()

    def __getattr__(self, k):
        # Work around PyCharm bug https://youtrack.jetbrains.com/issue/PY-4213
        if k == "_cp":
            return
        return ConfigGroup(self._cp[k])

    __getitem__ = __getattr__

    def _read_file(self, flo):
        self._cp.read_file(flo)


class ConfigGroup(object):
    def __init__(self, section):
        self.__dict__['_section'] = section

    def __dir__(self):
        return self.__dict__['_section'].keys()

    def __getattr__(self, k):
        return _val_xform(self.__dict__['_section'][k])

    __getitem__ = __getattr__

    def __setattr__(self, k, v):
        logger.info(str(self.__class__.__name__) + '.__setattr__({k}, ...)'.format(k=k))
        self.__dict__['_section'][k] = str(v)

    __setitem__ = __setattr__


def _name_xform(o):
    """transform names to lowercase, without symbols (except underscore)
    Any chars other than alphanumeric are converted to an underscore
    """
    return re.sub('\W', '_', o.lower())


def _val_xform(v):
    if v == 'True':
        return True
    if v == 'False':
        return False
    try:
        return int(v)
    except ValueError:
        pass
    return v


_default_config = Config()
_default_config._read_file(resource_stream(__name__, '_data/defaults.ini'))

global_config = copy(_default_config)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nose.plugins.attrib import attr

from hgvs.exceptions import HGVSUnsupportedOperationError
import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.variantmapper


@attr(tags=["fx"])
class Test_VariantLengths(unittest.TestCase):
    """test length_change method for all variant types and cases"""

    success_tests = [
        # class AAExt(Edit, recordtype.recordtype('AAExt', [('ref', None), ('alt', None), ('aaterm', None), ('length', None),
        ("NP_000040.1:p.Ter314Trpext*45", 45),
        ("NP_000040.1:p.Met1ext-12", 12),

        # class AAFs(Edit, recordtype.recordtype('AAFs', [('ref', None), ('alt', None), ('length', None), ('uncertain', False)])):
        # UNSUPPORTED: ("NP_000001.1:p.Gly25Trpfs*10", ),   # requires seq fetch for length :-(

        # class AARefAlt(Edit, recordtype.recordtype('AARefAlt', [('ref', None), ('alt', None), ('uncertain', False)])):
        ("NP_006149.2:p.(Glu396delinsLysLys)", 1),   # NM_006158.3:c.1186delGinsAAAA
        ("NP_006149.2:p.(Glu397del)", -1),  # NM_006158.3:n.1288_1290delGAG
        ("NP_006149.2:p.(Glu397dup)", 1),  # NM_006158.3:n.1288_1290delGAGinsGAGGAG

        # class AASub(AARefAlt):
        ("NP_006149.2:p.(Glu396Lys)", 0),  # NM_006158.3:c.1186delGinsA

        # class Conv(Edit, recordtype.recordtype('Conv', [('from_ac', None), ('from_type', None), ('from_pos', None),
        # UNSUPPORTED

        # class Dup(Edit, recordtype.recordtype('Dup', [('ref', None), ('uncertain', False)])):
        ("NM_006158.3:n.1291_1293dup", 3),
        ("NM_006158.3:n.1291_1293dupGAG", 3),
        
        # class Inv(Edit, recordtype.recordtype('Inv', [('ref', None), ('uncertain', False)])):
        ("NM_006158.3:n.1291_1293inv", 0),

        # class NACopy(Edit, recordtype.recordtype('NACopy', ['copy', ('uncertain', False)])):
        ("NC_000014.8:g.88401076_88459508copy4", (88459508-88401076+1) * 4),

        # class NADupN(Edit, recordtype.recordtype('NADupN', ['n', ('uncertain', False)])):
        ("NM_015702.2:c.307_324dup18", (324-307+1) * 18),

        # class    NARefAlt(Edit, recordtype.recordtype('NARefAlt', [('ref', None), ('alt', None), ('uncertain', False)])):
        ("NM_000314.4:c.706G>T", 0),
        ("NM_000314.4:c.706delG", -1),
        ("NM_000314.4:c.706delGinsT", 0),
        ("NM_000314.4:c.706_708del", -3),
        ("NM_000314.4:c.706_706delG", -1),
        ("NM_000314.4:c.706_708delGAC", -3),

        ("NM_000314.4:c.706_708delGACinsT", -2),
        ("NM_000314.4:c.706_708delGACinsTT", -1),
        ("NM_000314.4:c.706_708delGACinsTTG", 0),
        ("NM_000314.4:c.706_708delGACinsTTGT", +1),
        ("NM_000314.4:c.706_708del3insTTGT", +1),
        ("NM_000314.4:c.706_708del3", -3),
        ("NM_004655.3:c.-12_8delAGCTCCCTCACCATGAGTAG", -20),

        ("NM_000314.4:c.706_707insT", +1),
        ("NM_000314.4:c.706_707insTT", +2),
        #See #319:: ("NM_000249.3:c.1897-3del", ),
        #See #319:: ("NM_000249.3:c.1897-3_1897-1del", ),
        #See #319:: ("NM_000249.3:c.1897-3_1897-1delinsACGT", ),
        ]

    error_tests = [
        "NM_000249.3:c.1897_1897-3del844insGGAAA",
        "NM_003777.3:c.13552_*36del57",
        ]

    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect()
        cls.vm = hgvs.variantmapper.VariantMapper(cls.hdp)
        cls.hp = hgvs.parser.Parser()

    def test_success_cases(self):
        "posedit length_change: test supported variant types"
        for tv, tl in self.success_tests:
            v = self.hp.parse_hgvs_variant(tv)
            self.assertEqual(v.posedit.length_change(), tl)

    def test_error_cases_w_exceptions(self):
        "posedit length_change: raise exception on error"
        for tv in self.error_tests:
            v = self.hp.parse_hgvs_variant(tv)
            with self.assertRaises(HGVSUnsupportedOperationError):
                _ = v.posedit.length_change()

    def test_error_cases_w_error_values(self):
        "posedit length_change: return None on error"
        for tv in self.error_tests:
            v = self.hp.parse_hgvs_variant(tv)
            self.assertIsNone(v.posedit.length_change(on_error_raise=False))



# -*- coding: utf-8 -*-

"""Tests uta postgresql client"""


from __future__ import absolute_import, division, print_function, unicode_literals

import os
import re
import unittest

from hgvs.exceptions import HGVSDataNotAvailableError
import hgvs.dataproviders.uta
import hgvs.edit
import hgvs.location
import hgvs.posedit
import hgvs.variantmapper
import hgvs.variant


class UTA_Base(object):
    def test_get_acs_for_protein_seq(self):
        exp = ['NP_001005405.1', 'MD5_8fc09b1d9a38a8c55176a0fa922df227']
        s = """
        mgccgcsggc gsgcggcgsg sggcgsgcgg cgssccvpic cckpvcccvp acscsscgsc
        ggskggcgsc gsskggcgsc gcsqsncckp ccsssgcgsf ccqsscskpc ccqssccqss
        cckpcccqss ccqsscfkpc ccqssccvpv ccqcki
        """

        s = re.sub('\s+', '', s.upper())
        self.assertEqual(sorted(self.hdp.get_acs_for_protein_seq(s)), sorted(exp))

        exp = ['NP_071928.2', 'MD5_ffb0d4adbd5e0b5d71678228b3696984']
        s = """
        masetektha llqtcstesl isslglgafc lvadrllqfs tiqqndwlra lsdnavhcvi
        gmwswavvtg ikkktdfgei ilagflasvi dvdhfflags mslkaaltlp rrpflhcstv
        ipvvvltlkf tmhlfklkds wcflpwmlfi swtshhirdg irhglwicpf gktsplpfwl
        yviitsslph icsfvmyltg trqmmsskhg vridv
        """

        s = re.sub('\s+', '', s.upper())
        self.assertEqual(sorted(self.hdp.get_acs_for_protein_seq(s)), sorted(exp))

    def test_get_gene_info(self):
        gene_info = self.hdp.get_gene_info('VHL')
        self.assertEqual('VHL', gene_info['hgnc'])
        self.assertEqual('3p25.3', gene_info['maploc'])
        self.assertEqual(6, len(gene_info))

    def test_get_tx_exons(self):
        tx_exons = self.hdp.get_tx_exons('NM_000551.3', 'NC_000003.11', 'splign')
        self.assertEqual(3, len(tx_exons))

    def test_get_tx_exons_invalid_tx_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_exons('NM_999999.9', 'NC_000003.11', 'splign')

    def test_get_tx_exons_invalid_alt_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_exons('NM_000551.3', 'NC_000999.9', 'splign')

    def test_get_tx_exons_invalid_alt_aln_method(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_exons('NM_000551.3', 'NC_000999.9', 'best')

    def test_get_tx_for_gene(self):
        tig = self.hdp.get_tx_for_gene('VHL')
        self.assertEqual(12, len(tig))

    def test_get_tx_for_gene_invalid_gene(self):
        tig = self.hdp.get_tx_for_gene('GENE')
        self.assertEqual(0, len(tig))

    def test_get_tx_info(self):
        tx_info = self.hdp.get_tx_info('NM_000051.3', 'AC_000143.1', 'splign')
        self.assertEqual(385, tx_info['cds_start_i'])
        self.assertEqual(9556, tx_info['cds_end_i'])
        self.assertEqual('AC_000143.1', tx_info['alt_ac'])

    def test_get_tx_info_invalid_tx_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp.get_tx_info('NM_999999.9', 'AC_000143.1', 'splign')

    def test_get_tx_mapping_options(self):
        tx_mapping_options = self.hdp.get_tx_mapping_options('NM_000551.3')
        self.assertIn(['NM_000551.3', 'NC_000003.11', 'splign'], tx_mapping_options)
        self.assertIn(['NM_000551.3', 'NC_000003.11', 'blat'], tx_mapping_options)

    def test_get_tx_mapping_options_invalid(self):
        tx_info_options = self.hdp.get_tx_mapping_options('NM_999999.9')
        self.assertEqual(tx_info_options, [])

    def test_get_tx_seq(self):
        tsi = self.hdp._get_tx_seq('NM_000551.3')
        seq = tsi
        self.assertEqual(4560, len(seq))
        self.assertTrue(seq.startswith('cctcgcctccgttacaacgg'.upper()))
        self.assertTrue(seq.endswith('aaaggaaatag'.upper()))

    def test_get_tx_seq_invalid_ac(self):
        with self.assertRaises(HGVSDataNotAvailableError):
            self.hdp._get_tx_seq('NM_999999.9')


class Test_hgvs_dataproviders_uta_UTA_default(unittest.TestCase, UTA_Base):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect()


class Test_hgvs_dataproviders_uta_UTA_default_with_pooling(unittest.TestCase, UTA_Base):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(pooling=True)


class TestUTACache(Test_hgvs_dataproviders_uta_UTA_default):
    def _create_cdna_variant(self):
        start = hgvs.location.SimplePosition(118898437)
        end = hgvs.location.SimplePosition(118898437)
        iv = hgvs.location.Interval(start=start, end=end)
        edit = hgvs.edit.NARefAlt(ref='G', alt='T')
        posedit = hgvs.posedit.PosEdit(pos=iv, edit=edit)
        genomic_variant = hgvs.variant.SequenceVariant(ac='NC_000011.9', type='g', posedit=posedit, )
        variantmapper = hgvs.variantmapper.VariantMapper(self.hdp)
        return variantmapper.g_to_c(genomic_variant, 'NM_001164277.1')

    def test_deterministic_cache_results(self):
        """
        Check that identical request to the UTA yields the same results.
        """
        var1 = self._create_cdna_variant()
        var2 = self._create_cdna_variant()
        self.assertEqual(str(var1), str(var2))


if __name__ == '__main__':
    unittest.main()

    ## <LICENSE>
    ## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
    ## 
    ## Licensed under the Apache License, Version 2.0 (the "License");
    ## you may not use this file except in compliance with the License.
    ## You may obtain a copy of the License at
    ## 
    ##     http://www.apache.org/licenses/LICENSE-2.0
    ## 
    ## Unless required by applicable law or agreed to in writing, software
    ## distributed under the License is distributed on an "AS IS" BASIS,
    ## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    ## See the License for the specific language governing permissions and
    ## limitations under the License.
    ## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nose.plugins.attrib import attr

from hgvs.exceptions import HGVSUnsupportedOperationError
import hgvs.dataproviders.uta
import hgvs.variantmapper
import hgvs.parser
import hgvs.normalizer

hdp = hgvs.dataproviders.uta.connect()


@attr(tags=["normalization"])
class Test_HGVSNormalizer(unittest.TestCase):
    """Tests for normalizer"""

    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()
        cls.norm = hgvs.normalizer.Normalizer(hdp, shuffle_direction=3, cross_boundaries=True)
        cls.norm5 = hgvs.normalizer.Normalizer(hdp, shuffle_direction=5, cross_boundaries=True)
        cls.normc = hgvs.normalizer.Normalizer(hdp, shuffle_direction=3, cross_boundaries=False)
        cls.norm5c = hgvs.normalizer.Normalizer(hdp, shuffle_direction=5, cross_boundaries=False)

    def test_c_normalizer(self):
        """Test normalizer for variant type c."""
        #3' shuffling
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.18_24inv'))),
                         'NM_001166478.1:c.19_23invCTCTT')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.31del'))),
                         'NM_001166478.1:c.35delT')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.35_36insT'))),
                         'NM_001166478.1:c.35dupT')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.36_37insTC'))),
                         'NM_001166478.1:c.36_37dupCT')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.35_36dup'))),
                         'NM_001166478.1:c.36_37dupCT')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.2_7delinsTTTAGA'))),
                         'NM_001166478.1:c.3_4delGAinsTT')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.30_31insT'))),
                         'NM_001166478.1:c.35dupT')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.59delG'))),
                         'NM_001166478.1:c.61delG')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.36_37insTCTCTC'))),
                         'NM_001166478.1:c.36_37dup3')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_000720.2:c.1208_1209dup2"))),
                         "NM_000720.2:c.1208_1211dupGTGT")
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.14_15insT'))),
                         'NM_000051.3:c.15dupT')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant("NM_030813.5:c.1305_1307invGGG"))),
                         "NM_030813.5:c.*942=")

        #5' shuffling
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.18_24inv'))),
                         'NM_001166478.1:c.19_23invCTCTT')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.34del'))),
                         'NM_001166478.1:c.31delT')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.35_36insT'))),
                         'NM_001166478.1:c.31dupT')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.36_37insTC'))),
                         'NM_001166478.1:c.35_36dupTC')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.35_36dup'))),
                         'NM_001166478.1:c.35_36dupTC')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.2_7delinsTTTAGA'))),
                         'NM_001166478.1:c.3_4delGAinsTT')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.30_31insT'))),
                         'NM_001166478.1:c.31dupT')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.61delG'))),
                         'NM_001166478.1:c.59delG')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.36_37insTCTCTC'))),
                         'NM_001166478.1:c.35_36dup3')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.14_15insT'))),
                         'NM_000051.3:c.14dupT')

        #Around exon-intron boundary
        self.assertEqual(str(self.normc.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.59delG'))),
                         'NM_001166478.1:c.60delG')
        self.assertEqual(str(self.normc.normalize(self.hp.parse_hgvs_variant("NM_030813.5:c.1305_1307invGGG"))),
                         "NM_030813.5:c.1419=")
        self.assertEqual(str(self.norm5c.normalize(self.hp.parse_hgvs_variant('NM_001166478.1:c.61delG'))),
                         'NM_001166478.1:c.61delG')
        self.assertEqual(str(self.norm5c.normalize(self.hp.parse_hgvs_variant('NM_001110792.1:c.1030_1035del'))),
                         'NM_001110792.1:c.1029_1034delGAGCGG')
        self.assertRaises(HGVSUnsupportedOperationError, self.normc.normalize,
                          self.hp.parse_hgvs_variant('NM_001166478.1:c.59_61del'))

        #UTR variants
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.-5_-4insA'))),
                         'NM_000051.3:c.-3dupA')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.-4_-3insAC'))),
                         'NM_000051.3:c.-3_-2dupAC')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.-2_-1insCA'))),
                         'NM_000051.3:c.-1_1dupCA')

        self.assertEqual(str(self.normc.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.-2_-1insCA'))),
                         'NM_000051.3:c.-1_1insAC')

        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.-4_-3insA'))),
                         'NM_000051.3:c.-4dupA')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.1_2insCA'))),
                         'NM_000051.3:c.-1_1dupCA')

        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.*2_*3insT'))),
                         'NM_000051.3:c.*4dupT')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.9170_9171insAT'))),
                         'NM_000051.3:c.9171_*1dupAT')

        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.*4_*5insT'))),
                         'NM_000051.3:c.*3dupT')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NM_000051.3:c.9171_*1insA'))),
                         'NM_000051.3:c.9171dupA')

    def test_g_normalizer(self):
        """Test normalizer for variant type g."""
        #3' shuffling
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917122_49917123insA'))),
                         'NC_000006.11:g.49917127dupA')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917121_49917122insGA'))),
                         'NC_000006.11:g.49917122_49917123dupGA')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917122_49917123dup'))),
                         'NC_000006.11:g.49917122_49917123dupGA')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917122_49917123dupGA'))),
                         'NC_000006.11:g.49917122_49917123dupGA')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917098delC'))),
                         'NC_000006.11:g.49917099delC')
        self.assertEqual(str(self.norm.normalize(self.hp.parse_hgvs_variant(
            'NC_000006.11:g.49917151_49917156delinsTCTAAA'))), 'NC_000006.11:g.49917154_49917155delTCinsAA')

        #5' shuffling
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917122_49917123insA'))),
                         'NC_000006.11:g.49917123dupA')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917121_49917122insGA'))),
                         'NC_000006.11:g.49917121_49917122dupAG')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917122_49917123dup'))),
                         'NC_000006.11:g.49917121_49917122dupAG')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917122_49917123dupGA'))),
                         'NC_000006.11:g.49917121_49917122dupAG')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant('NC_000006.11:g.49917099delC'))),
                         'NC_000006.11:g.49917098delC')
        self.assertEqual(str(self.norm5.normalize(self.hp.parse_hgvs_variant(
            'NC_000006.11:g.49917151_49917156delinsTCTAAA'))), 'NC_000006.11:g.49917154_49917155delTCinsAA')


if __name__ == '__main__':
    unittest.main()

## <LICENSE>
## Copyright 2015 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-

"""hgvs is a package to parse, format, and manipulate biological sequence
variants.  See https://bitbucket.org/biocommons/hgvs/ for details.

Example use:

>>> import hgvs.dataproviders.uta
>>> import hgvs.parser
>>> import hgvs.variantmapper

# start with these variants as strings
>>> hgvs_g, hgvs_c = 'NC_000007.13:g.36561662C>T', 'NM_001637.3:c.1582G>A'

# parse the genomic variant into a Python structure
>>> hp = hgvs.parser.Parser()
>>> var_g = hp.parse_hgvs_variant(hgvs_g)
>>> var_g
SequenceVariant(ac=NC_000007.13, type=g, posedit=36561662C>T)

# SequenceVariants are composed of structured objects, e.g.,
>>> var_g.posedit.pos.start
SimplePosition(base=36561662, uncertain=False)

# format by stringification 
>>> str(var_g)
'NC_000007.13:g.36561662C>T'

# initialize the mapper for GRCh37 with splign-based alignments
>>> hdp = hgvs.dataproviders.uta.connect()
>>> evm = hgvs.variantmapper.EasyVariantMapper(hdp,
...          primary_assembly='GRCh37', alt_aln_method='splign',
...          replace_reference=True)

# identify transcripts that overlap this genomic variant
>>> transcripts = evm.relevant_transcripts(var_g)
>>> sorted(transcripts)
['NM_001177506.1', 'NM_001177507.1', 'NM_001637.3']

# map genomic variant to one of these transcripts
>>> var_c = evm.g_to_c(var_g, 'NM_001637.3')
>>> var_c
SequenceVariant(ac=NM_001637.3, type=c, posedit=1582G>A)
>>> str(var_c)
'NM_001637.3:c.1582G>A'

# CDS coordinates use BaseOffsetPosition to support intronic offsets
>>> var_c.posedit.pos.start
BaseOffsetPosition(base=1582, offset=0, datum=1, uncertain=False)

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import pkg_resources
import re
import warnings

from .config import global_config

logger = logging.getLogger(__name__)

_is_released_version = False


try:
    __version__ = pkg_resources.get_distribution('hgvs').version
    if re.match('^\d+\.\d+\.\d+$', __version__) is not None:
        _is_released_version = True
except pkg_resources.DistributionNotFound as e:
    warnings.warn("can't get __version__ because %s package isn't installed" % __package__, Warning)
    __version__ = None

# Make sure we're showing DeprecationWarnings
warnings.filterwarnings('default', '', DeprecationWarning)

logger.info("hgvs " + __version__ + "; released: " + str(_is_released_version))


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



############################################################################
## Project config

import hgvs
version = hgvs.__version__
release = str(hgvs.__version__)

project = u'HGVS'
authors = project + ' Contributors'
copyright = u'2015, ' + authors

extlinks = {
    'issue': ('https://bitbucket.org/biocommons/hgvs/issue/%s', 'issue '),
    }

man_pages = [
    ('index', project, project + u' Documentation', [project + u' Contributors'], 1)
]

# even for local builds (setup.py has dependency)
html_theme = 'sphinx_rtd_theme'


############################################################################
## Boilerplate

autodoc_default_flags = ['members', 'undoc-members', 'show-inheritance'] #, 'inherited-members']
autoclass_content = 'both'
exclude_patterns = ['build','static','templates','themes']
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.pngmath',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    ]
html_favicon = 'static/favicon.ico'
html_logo = 'static/hgvs-logo.png'
html_static_path = ['static']
html_title = '{project} {release}'.format(project = project, release = release)
intersphinx_mapping = {
    'http://docs.python.org/': None,
    }
master_doc = 'index'
pygments_style = 'sphinx'
source_suffix = '.rst'
templates_path = ['templates']

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



import re

import requests

from ..exceptions import HGVSDataNotAvailableError
from ..decorators import lru_cache

# TODO: Move sequence fetching to another package.
# TODO: Consider using eutils to get request throttling.
# N.B. >> used (rather than >>>) to not incur web requests during testing


def _fetch_seq_ensembl(ac, start_i=None, end_i=None):
    """Fetch the specified sequence slice from Ensembl using the public
    REST interface.

    An interbase interval may be optionally provided with start_i and
    end_i. However, the Ensembl REST interface does not currently
    accept intervals, so the entire sequence is returned and sliced
    locally.

    >> len(_fetch_seq_ensembl('ENSP00000288602'))
    766

    >> _fetch_seq_ensembl('ENSP00000288602',0,10)
    u'MAALSGGGGG'

    >> _fetch_seq_ensembl('ENSP00000288602')[0:10]
    u'MAALSGGGGG'

    >> _fetch_seq_ensembl('ENSP00000288602',0,10) == _fetch_seq_ensembl('ENSP00000288602')[0:10]
    True

    """

    url_fmt = "http://rest.ensembl.org/sequence/id/{ac}"
    url = url_fmt.format(ac=ac)
    r = requests.get(url, headers={"Content-Type": "application/json"})
    r.raise_for_status()
    seq = r.json()['seq']
    return seq if (start_i is None or end_i is None) else seq[start_i:end_i]


def _fetch_seq_ncbi(ac, start_i=None, end_i=None):
    """Fetch sequences from NCBI using the eutils interface.  

    An interbase interval may be optionally provided with start_i and
    end_i. NCBI eutils will return just the requested subsequence,
    which might greatly reduce payload sizes (especially with
    chromosome-scale sequences).

    >> len(_fetch_seq_ncbi('NP_056374.2'))
    1596

    Pass the desired interval rather than using Python's [] slice
    operator.

    >> _fetch_seq_ncbi('NP_056374.2',0,10)
    'MESRETLSSS'

    >> _fetch_seq_ncbi('NP_056374.2')[0:10]
    'MESRETLSSS'

    >> _fetch_seq_ncbi('NP_056374.2',0,10) == _fetch_seq_ncbi('NP_056374.2')[0:10]
    True

    """

    url_fmt = "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=nucleotide&id={ac}&rettype=fasta"
    if (start_i is None or end_i is None):
        url = url_fmt.format(ac=ac)
    else:
        url_fmt += "&seq_start={start}&seq_stop={stop}"
        url = url_fmt.format(ac=ac, start=start_i + 1, stop=end_i)
    resp = requests.get(url)
    resp.raise_for_status()
    return ''.join(resp.content.splitlines()[1:])


@lru_cache(maxsize=20)
def fetch_seq(ac, start_i=None, end_i=None):
    """return a subsequence of the given accession for the
    interbase interval [start_i,end_i)

    Fetches sequences from NCBI eutils and Ensembl REST interfaces
    (currently).  This class is primarily intended as a mixin for HGVS
    data providers that doen't otherwise have access to sequence data.

    Without an interval, the full sequence is returned::

    >> len(fetch_seq('NP_056374.2'))
    1596

    Therefore, it's preferable to provide the interval rather than
    using Python slicing sequence on the delivered sequence::

    >> fetch_seq('NP_056374.2',0,10)   # This!
    'MESRETLSSS'

    >> fetch_seq('NP_056374.2')[0:10]  # Not this!
    'MESRETLSSS'

    >> fetch_seq('NP_056374.2',0,10) == fetch_seq('NP_056374.2')[0:10]
    True

    Providing intervals is especially important for large sequences::

    >> fetch_seq('NC_000001.10',2000000,2000030)
    'ATCACACGTGCAGGAACCCTTTTCCAAAGG'

    This call will pull back 30 bases plus overhead; without the
    interval, one would receive 250MB of chr1 plus overhead.

    Essentially any RefSeq, Genbank, BIC, or Ensembl sequence may be
    fetched:

    >> [(ac,fetch_seq(ac,0,25))
    ... for ac in ['NG_032072.1', 'NW_003571030.1', 'NT_113901.1', 'NC_000001.10', 'NP_056374.2', 'GL000191.1', 'KB663603.1',
    ...            'ENST00000288602', 'ENSP00000288602']]
    [('NG_032072.1', 'AAAATTAAATTAAAATAAATAAAAA'),
     ('NW_003571030.1', 'TTGTGTGTTAGGGTGCTCTAAGCAA'),
     ('NT_113901.1', 'GAATTCCTCGTTCACACAGTTTCTT'),
     ('NC_000001.10', 'NNNNNNNNNNNNNNNNNNNNNNNNN'),
     ('NP_056374.2', 'MESRETLSSSRQRGGESDFLPVSSA'),
     ('GL000191.1', 'GATCCACCTGCCTCAGCCTCCCAGA'),
     ('KB663603.1', 'TTTATTTATTTTAGATACTTATCTC'),
     ('ENST00000288602', u'CGCCTCCCTTCCCCCTCCCCGCCCG'),
     ('ENSP00000288602', u'MAALSGGGGGGAEPGQALFNGDMEP')]

    >> fetch_seq('NM_9.9')
    Traceback (most recent call last):
       ...
    HGVSDataNotAvailableError: No sequence available for NM_9.9

    >> fetch_seq('QQ01234')
    Traceback (most recent call last):
       ...
    HGVSDataNotAvailableError: No fetcher for accessions like QQ01234

    """

    _ac_dispatch = [
        {'re': re.compile('^(?:AC|N[CGMPRTW])_|^[A-L]\w\d|^U\d'),
         'fetcher': _fetch_seq_ncbi},
        {'re': re.compile('^ENS[TP]\d+'),
         'fetcher': _fetch_seq_ensembl},
    ]

    for drec in _ac_dispatch:
        if drec['re'].match(ac):
            try:
                return drec['fetcher'](ac, start_i, end_i)
            except requests.HTTPError:
                raise HGVSDataNotAvailableError("No sequence available for {ac}".format(ac=ac))
    raise HGVSDataNotAvailableError("No fetcher for accessions like {}".format(ac))


class SeqFetcher(object):
    """This class is intended primarily as a mixin for HGVS data providers
    that doen't otherwise have access to sequence data.  It uses the
    fetch_seq() function in this module to fetch sequences from
    several sources; see that function for details.

    >> sf = SeqFetcher()

    >> sf.fetch_seq('NP_056374.2',0,10)
    'MESRETLSSS'

    """

    def fetch_seq(self, ac, start_i=None, end_i=None):
        """See fetch_seq() *function* in this module for details"""
        return fetch_seq(ac, start_i, end_i)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import logging

from Bio.Seq import Seq
from bioutils.accessions import primary_assembly_accessions
from bioutils.sequences import reverse_complement
import recordtype

from hgvs.exceptions import HGVSDataNotAvailableError, HGVSUnsupportedOperationError, HGVSInvalidVariantError
import hgvs
import hgvs.location
import hgvs.normalizer
import hgvs.posedit
import hgvs.transcriptmapper
import hgvs.utils.altseq_to_hgvsp as altseq_to_hgvsp
import hgvs.utils.altseqbuilder as altseqbuilder
import hgvs.variant

from hgvs.decorators.lru_cache import lru_cache

_logger = logging.getLogger(__name__)


class VariantMapper(object):
    """Maps SequenceVariant objects between g., n., r., c., and p. representations.

    g⟷{c,n,r} projections are similar in that c, n, and r variants
    may use intronic coordinates. There are two essential differences
    that distinguish the three types:

    * Sequence start: In n and r variants, position 1 is the sequence
      start; in c variants, 1 is the transcription start site.
    * Alphabet: In n and c variants, sequences are DNA; in
      r. variants, sequences are RNA.
    
    This differences are summarized in this diagram::

      g ----acgtatgcac--gtctagacgt----      ----acgtatgcac--gtctagacgt----      ----acgtatgcac--gtctagacgt----
            \         \/         /              \         \/         /              \         \/         /
      c      acgtATGCACGTCTAGacgt         n      acgtatgcacgtctagacgt         r      acguaugcacgucuagacgu   
                 1                               1                                   1
      p          MetHisValTer
    
    The g excerpt and exon structures are identical. The g⟷n
    transformation, which is the most basic, accounts for the offset
    of the aligned sequences (shown with "1") and the exon structure.
    The g⟷c transformation is akin to g⟷n transformation, but
    requires an addition offset to account for the translation start
    site (c.1).  The CDS in uppercase. The g⟷c transformation is
    akin to g⟷n transformation with a change of alphabet.

    Therefore, this this code uses g⟷n as the core transformation
    between genomic and c, n, and r variants: All c⟷g and r⟷g
    transformations use n⟷g after accounting for the above
    differences. For example, c_to_g accounts for the transcription
    start site offset, then calls n_to_g.

    All methods require and return objects of type
    :class:`hgvs.variant.SequenceVariant`.

    """

    def __init__(self, hdp):
        self.hdp = hdp

    # ############################################################################
    # g⟷n
    def g_to_n(self, var_g, tx_ac, alt_aln_method='splign'):
        """Given a parsed g. variant, return a n. variant on the specified
        transcript using the specified alignment method (default is
        'splign' from NCBI).

        :param hgvs.variant.SequenceVariant var_g: a variant object
        :param str tx_ac: a transcript accession (e.g., NM_012345.6 or ENST012345678)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`) using transcript (n.) coordinates
        :raises HGVSInvalidVariantError: if var_g is not of type 'g'

        """

        if not (var_g.type == 'g'):
            raise HGVSInvalidVariantError('Expected a g. variant; got ' + str(var_g))
        tm = self._fetch_TranscriptMapper(tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method)
        pos_n = tm.g_to_n(var_g.posedit.pos)
        edit_n = self._convert_edit_check_strand(tm.strand, var_g.posedit.edit)
        var_n = hgvs.variant.SequenceVariant(ac=tx_ac, type='n', posedit=hgvs.posedit.PosEdit(pos_n, edit_n))
        return var_n

    def n_to_g(self, var_n, alt_ac, alt_aln_method='splign'):
        """Given a parsed n. variant, return a g. variant on the specified
        transcript using the specified alignment method (default is
        'splign' from NCBI).

        :param hgvs.variant.SequenceVariant var_n: a variant object
        :param str alt_ac: a reference sequence accession (e.g., NC_000001.11)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_n is not of type 'n'

        """

        if not (var_n.type == 'n'):
            raise HGVSInvalidVariantError('Expected a n. variant; got ' + str(var_n))
        tm = self._fetch_TranscriptMapper(tx_ac=var_n.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method)
        pos_g = tm.n_to_g(var_n.posedit.pos)
        edit_g = self._convert_edit_check_strand(tm.strand, var_n.posedit.edit)
        var_g = hgvs.variant.SequenceVariant(ac=alt_ac, type='g', posedit=hgvs.posedit.PosEdit(pos_g, edit_g))
        return var_g

    # ############################################################################
    # g⟷c
    def g_to_c(self, var_g, tx_ac, alt_aln_method='splign'):
        """Given a parsed g. variant, return a c. variant on the specified
        transcript using the specified alignment method (default is
        'splign' from NCBI).

        :param hgvs.variant.SequenceVariant var_g: a variant object
        :param str tx_ac: a transcript accession (e.g., NM_012345.6 or ENST012345678)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`) using CDS coordinates
        :raises HGVSInvalidVariantError: if var_g is not of type 'g'

        """

        if not (var_g.type == 'g'):
            raise HGVSInvalidVariantError('Expected a g. variant; got ' + str(var_g))

        tm = self._fetch_TranscriptMapper(tx_ac=tx_ac, alt_ac=var_g.ac, alt_aln_method=alt_aln_method)
        pos_c = tm.g_to_c(var_g.posedit.pos)
        edit_c = self._convert_edit_check_strand(tm.strand, var_g.posedit.edit)
        var_c = hgvs.variant.SequenceVariant(ac=tx_ac, type='c', posedit=hgvs.posedit.PosEdit(pos_c, edit_c))
        return var_c

    def c_to_g(self, var_c, alt_ac, alt_aln_method='splign'):
        """Given a parsed c. variant, return a g. variant on the specified
        transcript using the specified alignment method (default is
        'splign' from NCBI).

        :param hgvs.variant.SequenceVariant var_c: a variant object
        :param str alt_ac: a reference sequence accession (e.g., NC_000001.11)
        :param str alt_aln_method: the alignment method; valid values depend on data source
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_c is not of type 'c'

        """

        if not (var_c.type == 'c'):
            raise HGVSInvalidVariantError('Expected a cDNA (c.); got ' + str(var_c))

        tm = self._fetch_TranscriptMapper(tx_ac=var_c.ac, alt_ac=alt_ac, alt_aln_method=alt_aln_method)

        pos_g = tm.c_to_g(var_c.posedit.pos)
        edit_g = self._convert_edit_check_strand(tm.strand, var_c.posedit.edit)

        var_g = hgvs.variant.SequenceVariant(ac=alt_ac, type='g', posedit=hgvs.posedit.PosEdit(pos_g, edit_g))
        return var_g

    # ############################################################################
    # c⟷n
    # TODO: Identify use case for this code
    def c_to_n(self, var_c):
        """Given a parsed c. variant, return a n. variant on the specified
        transcript using the specified alignment method (default is
        'transcript' indicating a self alignment).

        :param hgvs.variant.SequenceVariant var_c: a variant object
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_c is not of type 'c'

        """

        if not (var_c.type == 'c'):
            raise HGVSInvalidVariantError('Expected a cDNA (c.); got ' + str(var_c))
        tm = self._fetch_TranscriptMapper(tx_ac=var_c.ac, alt_ac=var_c.ac, alt_aln_method='transcript')
        pos_n = tm.c_to_n(var_c.posedit.pos)
        if isinstance(var_c.posedit.edit, hgvs.edit.NARefAlt) or isinstance(var_c.posedit.edit, hgvs.edit.Dup) or isinstance(var_c.posedit.edit, hgvs.edit.NADupN) or isinstance(var_c.posedit.edit, hgvs.edit.Inv):
            edit_n = copy.deepcopy(var_c.posedit.edit)
        else:
            raise HGVSUnsupportedOperationError('Only NARefAlt/Dup/NADupN/Inv types are currently implemented')
        var_n = hgvs.variant.SequenceVariant(ac=var_c.ac, type='n', posedit=hgvs.posedit.PosEdit(pos_n, edit_n))
        return var_n

    def n_to_c(self, var_n):
        """Given a parsed n. variant, return a c. variant on the specified
        transcript using the specified alignment method (default is
        'transcript' indicating a self alignment).

        :param hgvs.variant.SequenceVariant var_n: a variant object
        :returns: variant object (:class:`hgvs.variant.SequenceVariant`)
        :raises HGVSInvalidVariantError: if var_n is not of type 'n'

        """

        if not (var_n.type == 'n'):
            raise HGVSInvalidVariantError('Expected n. variant; got ' + str(var_n))
        tm = self._fetch_TranscriptMapper(tx_ac=var_n.ac, alt_ac=var_n.ac, alt_aln_method='transcript')
        pos_c = tm.n_to_c(var_n.posedit.pos)
        if isinstance(var_n.posedit.edit, hgvs.edit.NARefAlt) or isinstance(var_n.posedit.edit, hgvs.edit.Dup) or isinstance(var_n.posedit.edit, hgvs.edit.NADupN) or isinstance(var_n.posedit.edit, hgvs.edit.Inv):
            edit_c = copy.deepcopy(var_n.posedit.edit)
        else:
            raise HGVSUnsupportedOperationError('Only NARefAlt/Dup/NADupN/Inv types are currently implemented')
        var_c = hgvs.variant.SequenceVariant(ac=var_n.ac, type='c', posedit=hgvs.posedit.PosEdit(pos_c, edit_c))
        return var_c

    # ############################################################################
    # c ⟶ p
    # TODO: c_to_p needs refactoring
    def c_to_p(self, var_c, pro_ac=None):
        """
        Converts a c. SequenceVariant to a p. SequenceVariant on the specified protein accession
        Author: Rudy Rico

        :param SequenceVariant var_c: hgvsc tag
        :param str pro_ac: protein accession
        :rtype: hgvs.variant.SequenceVariant

        """

        class RefTranscriptData(recordtype.recordtype('RefTranscriptData',
                                                      ['transcript_sequence', 'aa_sequence', 'cds_start', 'cds_stop',
                                                       'protein_accession'])):
            @classmethod
            def setup_transcript_data(cls, hdp, tx_ac, pro_ac):
                """helper for generating RefTranscriptData from for c_to_p"""
                tx_info = hdp.get_tx_identity_info(var_c.ac)
                tx_seq = hdp.fetch_seq(tx_ac)

                if tx_info is None or tx_seq is None:
                    raise HGVSDataNotAvailableError("Missing transcript data for accession: {}".format(tx_ac))

                # use 1-based hgvs coords
                cds_start = tx_info['cds_start_i'] + 1
                cds_stop = tx_info['cds_end_i']

                # padding list so biopython won't complain during the conversion
                tx_seq_to_translate = tx_seq[cds_start - 1:cds_stop]
                if len(tx_seq_to_translate) % 3 != 0:
                    ''.join(list(tx_seq_to_translate).extend(['N'] * ((3 - len(tx_seq_to_translate) % 3) % 3)))

                tx_seq_cds = Seq(tx_seq_to_translate)
                protein_seq = str(tx_seq_cds.translate())

                if pro_ac is None:
                    # get_acs... will always return at least the MD5_ accession
                    pro_ac = (hdp.get_pro_ac_for_tx_ac(tx_ac) or hdp.get_acs_for_protein_seq(protein_seq)[0])

                transcript_data = RefTranscriptData(tx_seq, protein_seq, cds_start, cds_stop, pro_ac)

                return transcript_data

        if not (var_c.type == 'c'):
            raise HGVSInvalidVariantError('Expected a cDNA (c.); got ' + str(var_c))

        reference_data = RefTranscriptData.setup_transcript_data(self.hdp, var_c.ac, pro_ac)
        builder = altseqbuilder.AltSeqBuilder(var_c, reference_data)

        # TODO - handle case where you get 2+ alt sequences back; currently get list of 1 element
        # loop structure implemented to handle this, but doesn't really do anything currently.
        all_alt_data = builder.build_altseq()

        var_ps = []
        for alt_data in all_alt_data:
            builder = altseq_to_hgvsp.AltSeqToHgvsp(reference_data, alt_data)
            var_p = builder.build_hgvsp()
            var_ps.append(var_p)

        var_p = var_ps[0]

        return var_p

    ############################################################################
    ## Internal methods

    def _replace_reference(self, var):
        """fetch reference sequence for variant and update (in-place) if necessary"""

        if var.type not in 'cgmnr':
            raise HGVSUnsupportedOperationError("Can only update references for type c, g, m, n, r")

        if var.posedit.edit.type == 'ins':
            # insertions have no reference sequence (zero-width), so return as-is
            return var
        if var.posedit.edit.type == 'con':
            # conversions have no reference sequence (zero-width), so return as-is
            return var

        if ((isinstance(var.posedit.pos.start, hgvs.location.BaseOffsetPosition) and var.posedit.pos.start.offset != 0)
            or (isinstance(var.posedit.pos.end, hgvs.location.BaseOffsetPosition) and var.posedit.pos.end.offset != 0)):
            _logger.info("Can't update reference sequence for intronic variant {}".format(var))
            return var

        # For c. variants, we need coords on underlying sequences
        if var.type == 'c':
            tm = self._fetch_TranscriptMapper(tx_ac=var.ac, alt_ac=var.ac, alt_aln_method='transcript')
            pos = tm.c_to_n(var.posedit.pos)
        else:
            pos = var.posedit.pos
        seq = self.hdp.fetch_seq(var.ac, pos.start.base - 1, pos.end.base)

        edit = var.posedit.edit
        if edit.ref != seq:
            _logger.debug("Replaced reference sequence in {var} with {seq}".format(var=var, seq=seq))
            edit.ref = seq

        return var

    @lru_cache(maxsize=128)
    def _fetch_TranscriptMapper(self, tx_ac, alt_ac, alt_aln_method):
        """
        Get a new TranscriptMapper for the given transcript accession (ac),
        possibly caching the result.
        """
        return hgvs.transcriptmapper.TranscriptMapper(self.hdp,
                                                      tx_ac=tx_ac,
                                                      alt_ac=alt_ac,
                                                      alt_aln_method=alt_aln_method)

    @staticmethod
    def _convert_edit_check_strand(strand, edit_in):
        """
        Convert an edit from one type to another, based on the stand and type
        """
        if isinstance(edit_in, hgvs.edit.NARefAlt):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                try:
                    # if smells like an int, do nothing
                    # TODO: should use ref_n, right?
                    int(edit_in.ref)
                    ref = edit_in.ref
                except (ValueError, TypeError):
                    ref = reverse_complement(edit_in.ref)
                edit_out = hgvs.edit.NARefAlt(ref=ref, alt=reverse_complement(edit_in.alt), )
        elif isinstance(edit_in, hgvs.edit.Dup):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                edit_out = hgvs.edit.Dup(ref=reverse_complement(edit_in.ref))
        elif isinstance(edit_in, hgvs.edit.NADupN):
            edit_out = copy.deepcopy(edit_in)
        elif isinstance(edit_in, hgvs.edit.Inv):
            if strand == 1:
                edit_out = copy.deepcopy(edit_in)
            else:
                try:
                    int(edit_in.ref)
                    ref = edit_in.ref
                except (ValueError, TypeError):
                    ref = reverse_complement(edit_in.ref)
                edit_out = hgvs.edit.Inv(ref=ref)
        else:
            raise NotImplementedError('Only NARefAlt/Dup/NADupN/Inv types are currently implemented')
        return edit_out


class EasyVariantMapper(VariantMapper):
    """Provides simplified variant mapping for a single assembly and
    transcript-reference alignment method.
    
    EasyVariantMapper is instantiated with a primary_assembly and
    alt_aln_method. These enable the following conveniences over
    VariantMapper:

    * The primary assembly and alignment method are used to
      automatically select an appropriate chromosomal reference
      sequence when mapping from a transcript to a genome (i.e.,
      c_to_g(...) and n_to_g(...)).

    * A new method, relevant_trancripts(g_variant), returns a list of
      transcript accessions available for the specified variant. These
      accessions are candidates mapping from genomic to trancript
      coordinates (i.e., g_to_c(...) and g_to_n(...)).

    IMPORTANT: Callers should be prepared to catch HGVSError
    exceptions. These will be thrown whenever a transcript maps
    ambiguously to a chromosome, such as for pseudoautosomal region
    transcripts.

    """

    def __init__(self, hdp,
                 primary_assembly=hgvs.global_config.mapping.assembly,
                 alt_aln_method=hgvs.global_config.mapping.alt_aln_method,
                 replace_reference=hgvs.global_config.mapping.replace_reference,
                 normalize=hgvs.global_config.mapping.normalize,
                 ):
        """
        :param str primary_assembly: assembly name ('GRCh37')
        :param str alt_aln_method: genome-transcript alignment method ('splign', 'blat', 'genewise')
        :param bool replace_reference: replace reference (entails additional network access)
        :param bool normalize: normalize variants
        :raises HGVSError subclasses: for a variety of mapping and data lookup failures
        """

        super(EasyVariantMapper, self).__init__(hdp=hdp)
        self.primary_assembly = primary_assembly
        self.alt_aln_method = alt_aln_method
        self.primary_assembly_accessions = set(primary_assembly_accessions[primary_assembly])
        self.replace_reference = replace_reference
        self.normalize = normalize
        self._norm = None
        if self.normalize:
            self._norm = hgvs.normalizer.Normalizer(hdp,alt_aln_method=alt_aln_method)

    def g_to_c(self, var_g, tx_ac):
        var_out = super(EasyVariantMapper, self).g_to_c(var_g, tx_ac, alt_aln_method=self.alt_aln_method)
        if self.replace_reference:
            self._replace_reference(var_out)
        return self._maybe_normalize(var_out)

    def g_to_n(self, var_g, tx_ac):
        var_out = super(EasyVariantMapper, self).g_to_n(var_g, tx_ac, alt_aln_method=self.alt_aln_method)
        if self.replace_reference:
            self._replace_reference(var_out)
        return self._maybe_normalize(var_out)

    def c_to_g(self, var_c):
        alt_ac = self._alt_ac_for_tx_ac(var_c.ac)
        var_out = super(EasyVariantMapper, self).c_to_g(var_c, alt_ac, alt_aln_method=self.alt_aln_method)
        if self.replace_reference:
            self._replace_reference(var_out)
        return self._maybe_normalize(var_out)

    def n_to_g(self, var_n):
        alt_ac = self._alt_ac_for_tx_ac(var_n.ac)
        var_out = super(EasyVariantMapper, self).n_to_g(var_n, alt_ac, alt_aln_method=self.alt_aln_method)
        if self.replace_reference:
            self._replace_reference(var_out)
        return self._maybe_normalize(var_out)

    def c_to_n(self, var_c):
        var_out = super(EasyVariantMapper, self).c_to_n(var_c)
        if self.replace_reference:
            self._replace_reference(var_out)
        return self._maybe_normalize(var_out)

    def n_to_c(self, var_n):
        var_out = super(EasyVariantMapper, self).n_to_c(var_n)
        if self.replace_reference:
            self._replace_reference(var_out)
        return self._maybe_normalize(var_out)

    def c_to_p(self, var_c):
        var_out = super(EasyVariantMapper, self).c_to_p(var_c)
        return self._maybe_normalize(var_out)


    def relevant_transcripts(self, var_g):
        """return list of transcripts accessions (strings) for given variant,
        selected by genomic overlap"""
        tx = self.hdp.get_tx_for_region(var_g.ac, self.alt_aln_method, var_g.posedit.pos.start.base,
                                        var_g.posedit.pos.end.base)
        return [e['tx_ac'] for e in tx]

    def _alt_ac_for_tx_ac(self, tx_ac):
        """return chromosomal accession for given transcript accession (and
        the primary_assembly and aln_method setting used to
        instantiate this EasyVariantMapper)

        """
        alt_acs = [e['alt_ac'] for e in self.hdp.get_tx_mapping_options(tx_ac)
                   if e['alt_aln_method'] == self.alt_aln_method and e['alt_ac'] in self.primary_assembly_accessions]
        if len(alt_acs) > 1:
            raise HGVSError("Multiple chromosomal alignments for {tx_ac} in {pa}"
                            "using {am} (likely paralog or pseudoautosomal region)".format(
                                tx_ac=tx_ac,
                                pa=self.primary_assembly,
                                am=self.alt_aln_method))
        if len(alt_acs) == 0:
            raise HGVSDataNotAvailableError("No alignments for {tx_ac} in {pa} using {am}".format(
                tx_ac=tx_ac,
                pa=self.primary_assembly,
                am=self.alt_aln_method))
        return alt_acs[0]    # exactly one remains


    def _maybe_normalize(self, var):
        """normalize variant if requested, and ignore HGVSUnsupportedOperationError
        This is better than checking whether the variant is intronic because
        future UTAs will support LRG, which will enable checking intronic variants.
        """
        if self._norm is not None:
            try:
                return self._norm.normalize(var)
            except HGVSUnsupportedOperationError as e:
                _logger.warn(str(e) + "; returning unnormalized variant")
                # fall through to return unnormalized variant
        return var


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
"""
hgvs.location -- classes for dealing with the locations of HGVS variants

This module provides for Representing the location of variants in HGVS nomenclature, including:

  * integers and integer intervals (e.g., NC_012345.6:g.3403243_3403248A>C)
  * CDS positions and intervals (e.g., NM_01234.5:c.56+12_56+14delAC)
  * CDS stop coordinates (e.g., NM_01234.5:c.*13A>C)  

Classes:

  * :class:`SimplePosition` -- a simple integer
  * :class:`BaseOffsetPosition` -- a position with datum, base, and offset for c. and r. coordinates
  * :class:`AAPosition` -- an amino acid position (with AA)
  * :class:`Interval` -- an interval of Positions
"""

import recordtype

from bioutils.sequences import aa1_to_aa3

from hgvs.exceptions import HGVSError, HGVSUnsupportedOperationError


SEQ_START = 0
CDS_START = 1
CDS_END = 2


class SimplePosition(recordtype.recordtype('SimplePosition', field_names=[('base', None), ('uncertain', False)])):
    def __str__(self):
        self.validate()
        s = '?' if self.base is None else str(self.base)
        return '(' + s + ')' if self.uncertain else s

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.base in None

    def _set_uncertain(self):
        "mark this location as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    def validate(self):
        "raise AssertionError if instance variables are invalid; otherwise return True"
        assert self.base is None or self.base >= 1, self.__class__.__name__ + ': base must be >= 1'
        return True

    def __sub__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot substract coordinates of different representations"
        return lhs.base - rhs.base


class BaseOffsetPosition(recordtype.recordtype(
    'BaseOffsetPosition',
    field_names=[('base', None), ('offset', 0), ('datum', SEQ_START), ('uncertain', False)])):
    """
    Class for dealing with CDS coordinates in transcript variants.

    This class models CDS positions using a `base` coordinate, which is
    measured relative to a specified `datum` (CDS_START or CDS_END), and
    an `offset`, which is 0 for exonic positions and non-zero for intronic
    positions.  **Positions and offsets are 1-based**, with no 0, per the HGVS
    recommendations.  (If you're using this with UTA, be aware that UTA
    uses interbase coordinates.)

    +----------+------------+-------+---------+------------------------------------------+
    | hgvs     | datum      | base  | offset  | meaning                                  |
    +==========+============+=======+=========+==========================================+
    | r.55     | SEQ_START  |   55  |      0  | RNA position 55                          |
    +----------+------------+-------+---------+------------------------------------------+
    | c.55     | CDS_START  |   55  |      0  | CDS position 55                          |
    +----------+------------+-------+---------+------------------------------------------+
    | c.55     | CDS_START  |   55  |      0  | CDS position 55                          |
    +----------+------------+-------+---------+------------------------------------------+
    | c.55+1   | CDS_START  |   55  |      1  | intronic variant +1 from boundary        |
    +----------+------------+-------+---------+------------------------------------------+
    | c.-55    | CDS_START  |  -55  |      0  | 5' UTR variant, 55 nt upstream of ATG    |
    +----------+------------+-------+---------+------------------------------------------+
    | c.1      | CDS_START  |    1  |      0  |   start codon                            |
    +----------+------------+-------+---------+------------------------------------------+
    | c.1234   | CDS_START  | 1234  |      0  | stop codon (assuming CDS length is 1233) |
    +----------+------------+-------+---------+------------------------------------------+
    | c.*1     | CDS_END    |    0  |      1  | STOP + 1                                 |
    +----------+------------+-------+---------+------------------------------------------+
    | c.*55    | CDS_END    |    0  |     55  | 3' UTR variant, 55 nt after STOP         |
    +----------+------------+-------+---------+------------------------------------------+
    """

    def validate(self):
        "raise AssertionError if instance variables are invalid; otherwise return True"
        assert self.base is None or self.base != 0, 'BaseOffsetPosition base may not be 0'
        assert self.base is None or self.datum == CDS_START or self.base >= 1, 'BaseOffsetPosition base must be >=1 for datum = SEQ_START or CDS_END'
        return True

    def __str__(self):
        self.validate()
        base_str = ('?' if self.base is None else '*' + str(self.base) if self.datum == CDS_END else str(self.base))
        offset_str = ('+?' if self.offset is None else '' if self.offset == 0 else '%+d' % self.offset)
        pos = base_str + offset_str
        return '(' + pos + ')' if self.uncertain else pos

    def _set_uncertain(self):
        "mark this location as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.base is None or self.offset is None

    def __sub__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot substract coordinates of different representations"
        if lhs.datum != rhs.datum:
            raise HGVSUnsupportedOperationError("Interval length measured from different datums is ill-defined")
        if lhs.offset != 0 or rhs.offset != 0:
            raise HGVSUnsupportedOperationError("Interval length with intronic offsets is ill-defined")
        if lhs.base == rhs.base:
            return lhs.offset - rhs.offset
        straddles_zero = 1 if (lhs.base > 0 and rhs.base < 0) else 0
        return lhs.base - rhs.base - straddles_zero


class AAPosition(recordtype.recordtype('AAPosition', field_names=[('base', None), ('aa', None), ('uncertain', False)])):
    def validate(self):
        "raise AssertionError if instance variables are invalid; otherwise return True"
        assert self.base is None or self.base >= 1, 'AAPosition location must be >=1'
        assert len(self.aa) == 1, 'More than 1 AA associated with position'
        return True

    def __str__(self):
        self.validate()
        pos = '?' if self.base is None else str(self.base)
        aa = '?' if self.aa is None else aa1_to_aa3(self.aa)
        s = aa + pos
        return '(' + s + ')' if self.uncertain else s

    @property
    def pos(self):
        """return base, for backward compatibility"""
        return self.base

    def _set_uncertain(self):
        "mark this location as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.base is None or self.aa is None

    def __sub__(lhs, rhs):
        assert type(lhs) == type(rhs), "Cannot substract coordinates of different representations"
        return lhs.base - rhs.base


class Interval(recordtype.recordtype('Interval', field_names=['start', ('end', None), ('uncertain', False)])):
    def validate(self):
        "raise AssertionError if instance variables are invalid; otherwise return True"
        return True

    def __str__(self):
        self.validate()
        if self.end is None or self.start == self.end:
            return str(self.start)
        iv = str(self.start) + '_' + str(self.end)
        return '(' + iv + ')' if self.uncertain else iv

    def _set_uncertain(self):
        "mark this interval as uncertain and return reference to self; this is called during parsing (see hgvs.ometa)"
        self.uncertain = True
        return self

    def _length(self):
        return 1 if self.end is None else self.end - self.start + 1

    @property
    def is_uncertain(self):
        """return True if the position is marked uncertain or undefined"""
        return self.uncertain or self.start.is_uncertain or self.end.is_uncertain


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



from setuptools import setup, find_packages

with open('doc/description.txt') as f:
    long_description = f.read()


setup(
    license = 'Apache License 2.0 (http://www.apache.org/licenses/LICENSE-2.0)',
    long_description = long_description,
    use_scm_version = True,
    zip_safe = True,

    author = 'HGVS Contributors',
    author_email = 'reecehart+hgvs@gmail.com',
    description = """HGVS Parser, Formatter, Mapper, Validator""",
    name = "hgvs",
    package_data = {'hgvs': ['_data/*']},
    packages = find_packages(),
    scripts = ['sbin/hgvs-shell'],
    url = 'https://bitbucket.org/biocommons/hgvs',

    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python",
        "Topic :: Database :: Front-Ends",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
    ],

    keywords = [
        'bioinformatics',
        'computational biology',
        'genome variants',
        'genome variation',
        'genomic variants',
        'genomic variation',
        'genomics',
        'hgvs',
    ],

    install_requires = [
        'biopython',
        'bioutils',
        'configparser',
        'parsley',
        'psycopg2',
        'recordtype',
        'requests>=1.0.0',
    ],

    setup_requires = [
        'setuptools_scm',
        'nose',
        'sphinx',
        'sphinxcontrib-fulltoc>=1.1',
        'wheel',
    ],

    tests_require = [
        'coverage',
        'unicodecsv',
    ],
)

## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
"""
hgvs.hgvsvalidator
"""

from .exceptions import HGVSValidationError, HGVSUnsupportedOperationError

import hgvs
import hgvs.location
import hgvs.parser
import hgvs.variantmapper

BASE_RANGE_ERROR_MSG = 'base start position must be <= end position'
OFFSET_RANGE_ERROR_MSG = 'offset start must be <= end position'
INS_ERROR_MSG = 'insertion length must be 1'
DEL_ERROR_MSG = 'Length implied by coordinates ({span_len})  must equal sequence deletion length ({del_len})'
AC_ERROR_MSG = 'Accession is not present in BDI database'
SEQ_ERROR_MSG = 'Variant reference ({var_ref_seq}) does not agree with reference sequence ({ref_seq})'


# TODO: #249: redisign validation interface for greater flexibility

BASE_OFFSET_COORD_TYPES = 'cnr'
SIMPLE_COORD_TYPES = 'gmp'

class Validator(object):
    """invoke intrinsic and extrinsic validation"""

    def __init__(self, hdp):
        self._ivr = IntrinsicValidator()
        self._evr = ExtrinsicValidator(hdp)

    def validate(self, var):
        return self._ivr.validate(var) and self._evr.validate(var)


class IntrinsicValidator(object):
    """
    Attempts to determine if the HGVS name is internally consistent
    """

    def validate(self, var):
        assert isinstance(var, hgvs.variant.SequenceVariant), 'variant must be a parsed HGVS sequence variant object'
        self._start_lte_end(var)
        self._ins_length_is_one(var)
        self._del_length(var)
        return True

    def _start_lte_end(self, var):
        # TODO: why check var.posedit.pos? should set, right?
        if var.type in SIMPLE_COORD_TYPES and var.posedit.pos:
            if var.posedit.pos.start.base > var.posedit.pos.end.base:
                raise HGVSValidationError(BASE_RANGE_ERROR_MSG)
        if var.type in BASE_OFFSET_COORD_TYPES and var.posedit.pos:
            if var.posedit.pos.start.base > var.posedit.pos.end.base:
                raise HGVSValidationError(BASE_RANGE_ERROR_MSG)
            elif var.posedit.pos.start.base == var.posedit.pos.end.base:
                if var.posedit.pos.start.offset > var.posedit.pos.end.offset:
                    raise HGVSValidationError(OFFSET_RANGE_ERROR_MSG)
            if var.posedit.pos.start.datum > var.posedit.pos.end.datum:
                raise HGVSValidationError(BASE_RANGE_ERROR_MSG)
        return True

    def _ins_length_is_one(self, var):
        # TODO: why check var.posedit.pos? should set, right?
        if var.posedit.edit.type == 'ins':
            if var.type in SIMPLE_COORD_TYPES and var.posedit.pos:
                if (var.posedit.pos.end.base - var.posedit.pos.start.base) != 1:
                    raise HGVSValidationError(INS_ERROR_MSG)
            if var.type in BASE_OFFSET_COORD_TYPES and var.posedit.pos:
                if ((var.posedit.pos.end.base + var.posedit.pos.end.offset) -
                    (var.posedit.pos.start.base + var.posedit.pos.start.offset)) != 1:
                    raise HGVSValidationError(INS_ERROR_MSG)
            return True

    def _del_length(self, var):
        if var.posedit.edit.type in ['del', 'delins']:
            ref_len = var.posedit.edit.ref_n
            if ref_len is None:
                return True

            if var.type in BASE_OFFSET_COORD_TYPES:
                if var.posedit.pos.start.datum != hgvs.location.CDS_END and var.posedit.pos.end.datum == hgvs.location.CDS_END:
                    raise HGVSUnsupportedOperationError(
                        "Validating deletion length across CDS end is unsupported ({}); consider projecting to n. first".format(str(var)))
                if not ((var.posedit.pos.start.offset == var.posedit.pos.end.offset == 0) or
                        (var.posedit.pos.start.base == var.posedit.pos.end.base)):
                    raise HGVSUnsupportedOperationError(
                        "Validating deletion length for variants across exon/intron junctions is unsupported ({})".format(str(var)))
                span_len = ((var.posedit.pos.end.base + var.posedit.pos.end.offset) -
                            (var.posedit.pos.start.base + var.posedit.pos.start.offset) + 1)
                # BO numbering is missing 0 (... -2 -1, 1, 2 ...) ⇒ adjust length if crossing
                if var.posedit.pos.start.base < 0 and var.posedit.pos.end.base > 0:
                    span_len -= 1
            else:
                span_len = var.posedit.pos.end.base - var.posedit.pos.start.base + 1

            if span_len != ref_len:
                raise HGVSValidationError(DEL_ERROR_MSG.format(span_len=span_len, del_len=ref_len))
        return True


class ExtrinsicValidator():
    """
    Attempts to determine if the HGVS name validates against external data sources
    """

    def __init__(self, hdp):
        self.hdp = hdp
        self.vm = hgvs.variantmapper.VariantMapper(self.hdp)

    def validate(self, var):
        assert isinstance(var, hgvs.variant.SequenceVariant), 'variant must be a parsed HGVS sequence variant object'
        # TODO: #253: Add p. validation support
        if var.type == 'p':
            raise HGVSUnsupportedOperationError(
                "Validating p. reference sequences is unsupported ({}); see https://bitbucket.org/biocommons/hgvs/issues/253/ ".format(str(var)))
        self._ref_is_valid(var)
        return True

    def _ref_is_valid(self, var):
        # use reference sequence of original variant, even if later converted (eg c_to_n)
        if (var.type in BASE_OFFSET_COORD_TYPES
            and var.posedit.pos is not None
            and (var.posedit.pos.start.offset != 0 or var.posedit.pos.end.offset != 0)):
            raise HGVSUnsupportedOperationError(
                "Cannot validate sequence of an intronic variant ({})".format(str(var)))

        var_ref_seq = getattr(var.posedit.edit, 'ref', None)

        if var_ref_seq == '':
            var_ref_seq = None

        if var_ref_seq:
            # ref_seq is digit, as in 'del6'
            try:
                int(var_ref_seq)
                var_ref_seq = None
            except ValueError:
                pass

        if var_ref_seq is not None:
            var_x = self.vm.c_to_n(var) if var.type == 'c' else var
            ref_seq = self.hdp.fetch_seq(var_x.ac, var_x.posedit.pos.start.base - 1, var_x.posedit.pos.end.base)
            if ref_seq != var_ref_seq:
                raise HGVSValidationError(str(var) + ': ' + SEQ_ERROR_MSG.format(ref_seq=ref_seq,
                                                                                 var_ref_seq=var_ref_seq))

        return True



## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nose.plugins.attrib import attr

from hgvs.exceptions import HGVSError, HGVSUnsupportedOperationError
import hgvs.location


@attr(tags=["quick", "models"])
class Test_SimplePosition(unittest.TestCase):
    def test_success(self):
        self.assertEqual(str(hgvs.location.SimplePosition(5)), '5')
        self.assertEqual(str(hgvs.location.SimplePosition(5, uncertain=True)), '(5)')
        self.assertEqual(str(hgvs.location.SimplePosition(None)), '?')

    def test_failure(self):
        with self.assertRaises(AssertionError):
            self.assertEqual(hgvs.location.SimplePosition(-1), 'SHOULD FAIL')

    def test_simple_subtraction(self):
        self.assertEqual(
            hgvs.location.SimplePosition(5) - hgvs.location.SimplePosition(3),
            2)


@attr(tags=["quick"])
class Test_BaseOffsetPosition(unittest.TestCase):
    def test_success(self):
        # r.5
        cdsp = hgvs.location.BaseOffsetPosition(5)
        self.assertEqual(cdsp.datum, hgvs.location.SEQ_START)
        self.assertEqual(cdsp.base, 5)
        self.assertEqual(cdsp.offset, 0)
        self.assertEqual(str(cdsp), '5')

        #r.5+6
        cdsp.offset = 6
        self.assertEqual(str(cdsp), '5+6')

        #r.5+?
        cdsp.offset = None
        self.assertEqual(str(cdsp), '5+?')

        #r.(5+?)
        cdsp.uncertain = True
        self.assertEqual(str(cdsp), '(5+?)')

        # c.*5
        cdsp = hgvs.location.BaseOffsetPosition(5, datum=hgvs.location.CDS_END)
        self.assertEqual(cdsp.datum, hgvs.location.CDS_END)
        self.assertEqual(cdsp.base, 5)
        self.assertEqual(cdsp.offset, 0)
        self.assertEqual(str(cdsp), '*5')

        cdsp.uncertain = True
        self.assertEqual(str(cdsp), '(*5)')

        cdsp.offset = 7
        self.assertEqual(str(cdsp), '(*5+7)')


    def test_baseoffset_subtraction(self):
        v30 = hgvs.location.BaseOffsetPosition(3,0)
        v50 = hgvs.location.BaseOffsetPosition(5,0)
        v52 = hgvs.location.BaseOffsetPosition(5,2)
        v54 = hgvs.location.BaseOffsetPosition(5,4)

        self.assertEqual(v50-v30, 2)
        
        with self.assertRaises(HGVSError):
            _ = v54-v52
        with self.assertRaises(HGVSError):
            _ = v54-v30

        


@attr(tags=["quick"])
class Test_AAPosition(unittest.TestCase):
    def test_AAPosition(self):
        ap = hgvs.location.AAPosition(15, 'S')
        self.assertEqual(ap.pos, 15)
        self.assertEqual(str(ap), 'Ser15')

    def test_aaposition_subtraction(self):
        l1 = hgvs.location.AAPosition(15, 'S')
        l2 = hgvs.location.AAPosition(20, 'S')
        self.assertEqual(l2-l1, 5)


@attr(tags=["quick"])
class Test_Interval(unittest.TestCase):
    def test_Interval(self):
        ival = hgvs.location.Interval(hgvs.location.BaseOffsetPosition(base=12,
                                                                       offset=+34),
                                      hgvs.location.BaseOffsetPosition(base=56,
                                                                       offset=-78))
        self.assertEqual(ival.start.base, 12)
        self.assertEqual(ival.start.offset, 34)
        self.assertEqual(ival.end.base, 56)
        self.assertEqual(ival.end.offset, -78)
        self.assertEqual(str(ival), '12+34_56-78')

    def test_length(self):
        ival = hgvs.location.Interval(hgvs.location.BaseOffsetPosition(base=12,
                                                                       offset=0),
                                      hgvs.location.BaseOffsetPosition(base=50,
                                                                       offset=0))
        self.assertEqual(ival._length(), 39)


if __name__ == '__main__':
    unittest.main()


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from nose.plugins.attrib import attr

import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.variantmapper


@attr(tags=["quick"])
class Test_VariantMapper(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect()
        cls.vm = hgvs.variantmapper.VariantMapper(cls.hdp)
        cls.hp = hgvs.parser.Parser()

    def test_VariantMapper_quick(self):
        # From garcia.tsv:
        hgvs_g = 'NC_000007.13:g.36561662C>T'
        hgvs_c = 'NM_001637.3:c.1582G>A'
        hgvs_p = 'NP_001628.1:p.(Gly528Arg)'    # from Mutalyzer

        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        var_c = self.vm.g_to_c(var_g, 'NM_001637.3')
        var_p = self.vm.c_to_p(var_c)

        self.assertEqual(str(var_c), hgvs_c)
        self.assertEqual(str(var_p), hgvs_p)


@attr(tags=["quick"])
class Test_VariantMapper_Exceptions(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect()
        cls.vm = hgvs.variantmapper.VariantMapper(cls.hdp)
        cls.hp = hgvs.parser.Parser()

    def test_gcrp_invalid_input_type(self):
        hgvs_g = 'NC_000007.13:g.36561662C>T'
        hgvs_c = 'NM_001637.3:c.1582G>A'

        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        var_c = self.hp.parse_hgvs_variant(hgvs_c)

        cases = {
            'gc': (self.vm.g_to_c, (var_c, 'NM_001637.3')),
            'gr': (self.vm.g_to_n, (var_c, 'NM_001637.3')),
            'rg': (self.vm.n_to_g, (var_c, 'NM_001637.3')),
            'cg': (self.vm.c_to_g, (var_g, 'NM_001637.3')),
            'cr': (self.vm.c_to_n, (var_g, )),
            'rc': (self.vm.n_to_c, (var_g, )),
            'cp': (self.vm.c_to_p, (var_g, None)),
        }

        failures = []
        for key in cases:
            try:
                func, args = cases[key]
                var_result = func(*args)
                failures.append(key)
            except hgvs.exceptions.HGVSInvalidVariantError:
                pass

        self.assertFalse(failures, "conversions not failing: {}".format(failures))


    def test_gc_invalid_input_nm_accession(self):
        hgvs_g = 'NC_000007.13:g.36561662C>T'
        var_g = self.hp.parse_hgvs_variant(hgvs_g)
        with self.assertRaises(hgvs.exceptions.HGVSError):
            var_p = self.vm.c_to_p(var_g, 'NM_999999.1')

    def test_undefined_cds(self):
        """Raise exception when requesting mapping to/from c. with non-coding transcript"""
        hgvs_n = 'NR_111984.1:n.44G>A'  # legit
        hgvs_c = 'NR_111984.1:c.44G>A'  # bogus: c. with non-coding tx accession
        var_n = self.hp.parse_hgvs_variant(hgvs_n)
        var_c = self.hp.parse_hgvs_variant(hgvs_c)
        tx_ac = var_n.ac

        with self.assertRaises(hgvs.exceptions.HGVSUsageError):
            var_c = self.vm.n_to_c(var_n)  # n_to_c: transcript is non-coding

        with self.assertRaises(hgvs.exceptions.HGVSUsageError):
            var_c = self.vm.c_to_n(var_c)  # c_to_n: var_c is bogus


class Test_RefReplacement(unittest.TestCase):
    test_cases = [
        # These casese attempt to test reference update in four dimensions:
        # - variant type: n, c, g
        # - major mapping paths: c<->n, c<->g, n<->g
        # - variant class: sub, del, ins, delins, dup
        # - strand: +/-

        # ADRB2    │ NM_000024.5 │  239 │ 1481 │ NC_000005.9  │  1 │ 148206155,148208197 | 284=1X32=1X1724=
        # cseq = hdp.fetch_seq('NM_000024.5')
        # gseq = hdp.fetch_seq('NC_000005.9',148206155,148208197)
        # cseq[280:290] = 'CAATAGAAGC'
        # gseq[280:290] = 'CAATGGAAGC'
        #                      ^ @ n.285
        # These variants are in and around the first sub: 
        {u'c': 'NM_000024.5:c.42C>N',  u'g': 'NC_000005.9:g.148206436C>N',  u'n': u'NM_000024.5:n.281C>N'},
        {u'c': 'NM_000024.5:c.43A>N',  u'g': 'NC_000005.9:g.148206437A>N',  u'n': u'NM_000024.5:n.282A>N'},
        {u'c': 'NM_000024.5:c.44A>N',  u'g': 'NC_000005.9:g.148206438A>N',  u'n': u'NM_000024.5:n.283A>N'},
        {u'c': 'NM_000024.5:c.45T>N',  u'g': 'NC_000005.9:g.148206439T>N',  u'n': u'NM_000024.5:n.284T>N'},
        {u'c': 'NM_000024.5:c.46A>N',  u'g': 'NC_000005.9:g.148206440G>N',  u'n': u'NM_000024.5:n.285A>N'}, # ref repl
        {u'c': 'NM_000024.5:c.47G>N',  u'g': 'NC_000005.9:g.148206441G>N',  u'n': u'NM_000024.5:n.286G>N'},
        {u'c': 'NM_000024.5:c.48A>N',  u'g': 'NC_000005.9:g.148206442A>N',  u'n': u'NM_000024.5:n.287A>N'},
        {u'c': 'NM_000024.5:c.49A>N',  u'g': 'NC_000005.9:g.148206443A>N',  u'n': u'NM_000024.5:n.288A>N'},
        {u'c': 'NM_000024.5:c.50G>N',  u'g': 'NC_000005.9:g.148206444G>N',  u'n': u'NM_000024.5:n.289G>N'},
        {u'c': 'NM_000024.5:c.51C>N',  u'g': 'NC_000005.9:g.148206445C>N',  u'n': u'NM_000024.5:n.290C>N'},

        # ins, del, delins, dup:
        {u'c': 'NM_000024.5:c.46_47insNN',  u'g': 'NC_000005.9:g.148206440_148206441insNN',  u'n': 'NM_000024.5:n.285_286insNN'},
        {u'c': 'NM_000024.5:c.45_47delTAG',  u'g': 'NC_000005.9:g.148206439_148206441delTGG',  u'n': 'NM_000024.5:n.284_286delTAG'},
        {u'c': 'NM_000024.5:c.45_47delTAGinsNNNN',  u'g': 'NC_000005.9:g.148206439_148206441delTGGinsNNNN',  u'n': 'NM_000024.5:n.284_286delTAGinsNNNN'},
        {u'c': 'NM_000024.5:c.45_47delTAGinsNNNN',  u'g': 'NC_000005.9:g.148206439_148206441delTGGinsNNNN',  u'n': 'NM_000024.5:n.284_286delTAGinsNNNN'},
        {u'c': 'NM_000024.5:c.46dupA',  u'g': 'NC_000005.9:g.148206440dupG',  u'n': 'NM_000024.5:n.285dupA'},


        # IFNA16   │ NM_002173.2 │    6 │  576 │ NC_000009.11 │ -1 │  21216371, 21217310 | 691=2X246=
        # cseq = hdp.fetch_seq('NM_002173.2')
        # gseq = reverse_complement(hdp.fetch_seq('NC_000009.11',21216371,21217310))
        # cseq[685:695] = 'AAATTTCAAA'
        # gseq[685:695] = 'AAATTTTCAA'
        #                        ^^ @ n.692_693
        # These variants are in and around the 2X substitution
        {u'c': 'NM_002173.2:c.*110A>N',  u'g': 'NC_000009.11:g.21216625T>N',  u'n': u'NM_002173.2:n.686A>N'},
        {u'c': 'NM_002173.2:c.*111A>N',  u'g': 'NC_000009.11:g.21216624T>N',  u'n': u'NM_002173.2:n.687A>N'},
        {u'c': 'NM_002173.2:c.*112A>N',  u'g': 'NC_000009.11:g.21216623T>N',  u'n': u'NM_002173.2:n.688A>N'},
        {u'c': 'NM_002173.2:c.*113T>N',  u'g': 'NC_000009.11:g.21216622A>N',  u'n': u'NM_002173.2:n.689T>N'},
        {u'c': 'NM_002173.2:c.*114T>N',  u'g': 'NC_000009.11:g.21216621A>N',  u'n': u'NM_002173.2:n.690T>N'},
        {u'c': 'NM_002173.2:c.*115T>N',  u'g': 'NC_000009.11:g.21216620A>N',  u'n': u'NM_002173.2:n.691T>N'},
        {u'c': 'NM_002173.2:c.*116C>N',  u'g': 'NC_000009.11:g.21216619A>N',  u'n': u'NM_002173.2:n.692C>N'}, # ref repl
        {u'c': 'NM_002173.2:c.*117A>N',  u'g': 'NC_000009.11:g.21216618G>N',  u'n': u'NM_002173.2:n.693A>N'}, # ref repl
        {u'c': 'NM_002173.2:c.*118A>N',  u'g': 'NC_000009.11:g.21216617T>N',  u'n': u'NM_002173.2:n.694A>N'},
        {u'c': 'NM_002173.2:c.*119A>N',  u'g': 'NC_000009.11:g.21216616T>N',  u'n': u'NM_002173.2:n.695A>N'},

        # ins, del, delins, dup:
        {u'c': 'NM_002173.2:c.*115_*117insNN',  u'g': 'NC_000009.11:g.21216618_21216620insNN',  u'n': 'NM_002173.2:n.691_693insNN'},
        {u'c': 'NM_002173.2:c.*114_*117delTTCA',  u'g': 'NC_000009.11:g.21216618_21216621delGAAA',  u'n': 'NM_002173.2:n.690_693delTTCA'},
        {u'c': 'NM_002173.2:c.*115_*117delTCAinsNN',  u'g': 'NC_000009.11:g.21216618_21216620delGAAinsNN',  u'n': 'NM_002173.2:n.691_693delTCAinsNN'},
        {u'c': 'NM_002173.2:c.*115_*117delTCAinsNN',  u'g': 'NC_000009.11:g.21216618_21216620delGAAinsNN',  u'n': 'NM_002173.2:n.691_693delTCAinsNN'},
        {u'c': 'NM_002173.2:c.*115_*117dupTCA',  u'g': 'NC_000009.11:g.21216618_21216620dupGAA',  u'n': 'NM_002173.2:n.691_693dupTCA'},
    ]


    @classmethod
    def setUpClass(cls):
        def _parse_rec(rec):
            rec['pv'] = {x: cls.hp.parse_hgvs_variant(rec[x]) for x in 'cgn'}
            return rec

        cls.hdp = hgvs.dataproviders.uta.connect()
        cls.evm = hgvs.variantmapper.EasyVariantMapper(cls.hdp, replace_reference=True, primary_assembly='GRCh37', alt_aln_method='splign')
        cls.hp = hgvs.parser.Parser()
        cls.tests = [_parse_rec(rec) for rec in cls.test_cases]

    def test_replace_reference_sequence(self):
        """EasyVariantMapper: Replace invalid reference sequence"""

        for rec in self.tests:
            for x in 'cgn':
                pv = rec['pv'][x]
                if pv.posedit.edit.ref:
                    # replace ref with junk
                    pv.posedit.edit.ref = 'NNNNNN'
                self.evm._replace_reference(pv)
                self.assertEqual(rec[x], str(pv))


@attr(tags=["quick"])
class Test_EasyVariantMapper(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        hdp = hgvs.dataproviders.uta.connect()
        cls.hp = hgvs.parser.Parser()
        cls.evm = hgvs.variantmapper.EasyVariantMapper(hdp, primary_assembly='GRCh37', alt_aln_method='splign')
    
    def _test_mapping(self, hgvs_set):
        """given list of variant strings, test all valid combinations of
        g<->n<->c<->p mappings

        """
        parsed_variants = [(hv,self.hp.parse_hgvs_variant(hv)) for hv in hgvs_set]
        hgvs = {v.type: hv for hv,v in parsed_variants}
        pvs  = {v.type:  v for hv,v in parsed_variants}
    
        if 'g' in pvs and 'c' in pvs:
            self.assertEqual(hgvs['g'], str(self.evm.c_to_g(pvs['c'])))
            self.assertEqual(hgvs['c'], str(self.evm.g_to_c(pvs['g'], pvs['c'].ac)))
        if 'g' in pvs and 'n' in pvs:
            self.assertEqual(hgvs['g'], str(self.evm.n_to_g(pvs['n'])))
            self.assertEqual(hgvs['n'], str(self.evm.g_to_n(pvs['g'], pvs['n'].ac)))
        if 'c' in pvs and 'n' in pvs:
            self.assertEqual(hgvs['n'], str(self.evm.c_to_n(pvs['c'])))
            self.assertEqual(hgvs['c'], str(self.evm.n_to_c(pvs['n'])))
        if 'c' in pvs and 'p' in pvs:
            self.assertEqual(hgvs['p'], str(self.evm.c_to_p(pvs['c'])))

    def test_SNV(self):
        """EasyVariantMapper: smoketest with SNVs"""
        hgvs_set = ['NC_000007.13:g.36561662C>T', 'NM_001637.3:c.1582G>A', 'NM_001637.3:n.1983G>A', 'NP_001628.1:p.(Gly528Arg)']
        self._test_mapping(hgvs_set)

    def test_intronic(self):
        """EasyVariantMapper: smoketest with intronic SNVs"""
        hgvs_set = ['NC_000010.10:g.89711873A>C', 'NM_000314.4:c.493-2A>C', 'NM_000314.4:n.1524-2A>C', 'NP_000305.3:p.?']
        self._test_mapping(hgvs_set)


if __name__ == '__main__':
    unittest.main()


## <LICENSE>
## Copyright 2014 HGVS Contributors (https://bitbucket.org/biocommons/hgvs)
## 
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
## 
##     http://www.apache.org/licenses/LICENSE-2.0
## 
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
## </LICENSE>
